import warnings
from contextlib import contextmanager


_STRUCT_CONTEXT_STACK = []


@contextmanager
def _struct_context(object):
    _STRUCT_CONTEXT_STACK.append(object)
    try:
        yield None
    finally:
        _STRUCT_CONTEXT_STACK.pop(-1)


def unsafe():
    warnings.warn("struct.unsafe() is deprecated. Use map() with new_type argument to avoid validation.")
    return _struct_context('unsafe')


def _unsafe():
    return _struct_context('unsafe')


def skip_validate():
    return 'unsafe' in _STRUCT_CONTEXT_STACK



from collections import namedtuple

from phi.backend.dynamic_backend import DYNAMIC_BACKEND as math


SolveResult = namedtuple('SolveResult', ['iterations', 'x', 'residual'])


def broyden(function, x0, inv_J0, accuracy=1e-5, max_iterations=1000, back_prop=False):
    """
    Broyden's method for finding a root of the given function.
    Given a function `f: R^n -> R^n`, it finds an x such that `f(x)=0` within the specified `accuracy`.

    Boryden's method does not require explicit computations of the Jacobian except for the initial inverse `inv_J0`.

    :param function: Differentiable black-box function mapping from tensors like x0 to tensors like y
    :param x0: initial guess for x
    :param inv_J0: Approximation of the inverse Jacobian matrix of f at x0. The closer this is to the true matrix, the fewer iterations will be required.
    :param max_iterations: (optional) maximum number of CG iterations to perform
    :param back_prop: Whether to enable auto-differentiation. This induces a memory cost scaling with the number of iterations. Otherwise, the memory cost is constant.
    :param accuracy: (optional) the algorithm terminates once |f(x)| ≤ accuracy for every entry. If None, the algorithm runs until `max_iterations` is reached.
    :return: list of SolveResults with [0] being the forward solve result, [1] backward solve result (will be added once backward pass is computed)
    :rtype: SolveResult
    """
    x0 = math.to_float(x0)
    y0 = function(x0)
    inv_J0 = math.to_float(inv_J0)

    def broyden_loop(x, y, inv_J, iterations):
        # --- Adjust our guess for x ---
        dx = - math.einsum('bij,bj->bi', inv_J, y)  # - J^-1 * y
        next_x = x + dx
        next_y = function(next_x)
        df = next_y - y
        dy_back_projected = math.einsum('bij,bj->bi', inv_J, df)
        # --- Approximate next inverted Jacobian ---
        numerator = math.einsum('bi,bj,bjk->bik', dx - dy_back_projected, dx, inv_J)  # (dx - J^-1 * df) * dx^T * J^-1
        denominator = math.einsum('bi,bi->b', dx, dy_back_projected)  # dx^T * J^-1 * df
        next_inv_J = inv_J + numerator / denominator
        return [next_x, next_y, next_inv_J, iterations + 1]

    x_, y_, _, iterations = math.while_loop(_max_residual_condition(1, accuracy), broyden_loop, [x0, y0, inv_J0, 0], back_prop=back_prop, name='Broyden', maximum_iterations=max_iterations)
    return SolveResult(iterations, x_, y_)


def conjugate_gradient(function, y, x0, accuracy=1e-5, max_iterations=1000, back_prop=False):
    """
    Solve the linear system of equations `A·x=y`  using the conjugate gradient (CG) algorithm.
    A, x and y can have arbitrary matching shapes, i.e. this method can be used to solve vector and matrix equations.

    Since representing the matrix A in memory might not be feasible, a linear function of x can be specified that computes `function(x) = Ax`.

    The implementation is based on https://nvlpubs.nist.gov/nistpubs/jres/049/jresv49n6p409_A1b.pdf

    :param y: Desired output of `f(x)`
    :param function: linear function of x that returns A·x
    :param x0: initial guess for the value of x
    :param accuracy: (optional) the algorithm terminates once |f(x)-y| ≤ accuracy for every entry. If None, the algorithm runs until `max_iterations` is reached.
    :param max_iterations: (optional) maximum number of CG iterations to perform
    :param back_prop: Whether to enable auto-differentiation. This induces a memory cost scaling with the number of iterations. Otherwise, the memory cost is constant.
    :return: Pair containing the result for x and the number of iterations performed
    """
    y = math.to_float(y)
    x0 = math.to_float(x0)
    dx0 = residual0 = y - function(x0)
    dy0 = function(dx0)
    non_batch_dims = tuple(range(1, len(y.shape)))

    def cg_loop(x, dx, dy, residual, iterations):
        dx_dy = math.sum(dx * dy, axis=non_batch_dims, keepdims=True)
        step_size = math.divide_no_nan(math.sum(dx * residual, axis=non_batch_dims, keepdims=True), dx_dy)
        x += step_size * dx
        residual -= step_size * dy
        dx = residual - math.divide_no_nan(math.sum(residual * dy, axis=non_batch_dims, keepdims=True) * dx, dx_dy)
        dy = function(dx)
        return [x, dx, dy, residual, iterations + 1]

    x_, _, _, residual_, iterations_ = math.while_loop(_max_residual_condition(3, accuracy), cg_loop, [x0, dx0, dy0, residual0, 0], back_prop=back_prop, name="ConjGrad", maximum_iterations=max_iterations)
    return SolveResult(iterations_, x_, residual_)


def _max_residual_condition(residual_index, accuracy):
    """continue if the maximum deviation from zero is bigger than desired accuracy"""
    if accuracy is None:
        return lambda *args: True
    else:
        return lambda *args: math.max(math.abs(args[residual_index])) > accuracy



# pylint: disable-msg = wildcard-import, unused-wildcard-import, unused-import

from .physics.physics import *
from .physics.world import *
from .physics.schroedinger import *
from .physics.fluid import *
from .physics.burgers import *
from .physics.heat import *
from .physics.worldutil import *
from .physics.field import *
from .physics.obstacle import *
from .physics.material import *
from .physics.domain import *
from .physics.field.effect import *
from .physics.pressuresolver.solver_api import PoissonDomain, PoissonSolver
from .physics.pressuresolver.sparse import SparseCG, SparseSciPy
from .physics.pressuresolver.geom import GeometricCG
from .physics.pressuresolver.fourier import FourierSolver

from .data.fluidformat import *
from .data.dataset import *
from .data.stream import *
from .data.reader import *

from phi.geom import *
from phi import math, struct

from .viz import display
from .viz.display import show
from .app import *

import numpy

np = numpy

physics_config = GLOBAL_AXIS_ORDER



from phi import math
from . import Physics
from .field import CenteredGrid


class KuramotoSivashinsky(Physics):

    def __init__(self):
        Physics.__init__(self)

    def step(self, u, dt=1.0, **dependent_states):
        assert isinstance(u, CenteredGrid)
        grad = u.gradient()
        laplace = u.laplace()
        laplace2 = laplace.laplace()
        du_dt = -laplace - laplace2 - 0.5 * grad ** 2
        result = u + dt * du_dt
        result -= math.mean(result.data, axis=tuple(range(1, len(math.staticshape(result.data)))), keepdims=True)
        return result.copied_with(age=u.age + dt, name=u.name)



# coding=utf-8
import itertools

import numpy as np
from numpy import pi
from phi import math, struct
from phi.geom import AABox
from phi.physics.field import ConstantField, StaggeredGrid

from .field import Field, StaggeredSamplePoints
from .grid import CenteredGrid


def diffuse(field, amount, substeps=1):
    u"""
Simulate a finite-time diffusion process of the form dF/dt = α · ΔF on a given `Field` F with diffusion coefficient α.

If `field` is periodic (set via `extrapolation='periodic'`), diffusion may be simulated in Fourier space.
Otherwise, finite differencing is used to approximate the
    :param field: CenteredGrid, StaggeredGrid or ConstantField
    :param amount: number of Field, typically α · dt
    :param substeps: number of iterations to use
    :return: Field of same type as `field`
    :rtype: Field
    """
    if isinstance(field, ConstantField):
        return field
    if isinstance(field, StaggeredGrid):
        return struct.map(lambda grid: diffuse(grid, amount, substeps=substeps), field, leaf_condition=lambda x: isinstance(x, CenteredGrid))
    assert isinstance(field, CenteredGrid), "Cannot diffuse field of type '%s'" % type(field)
    if field.extrapolation == 'periodic' and not isinstance(amount, Field):
        fft_laplace = -(2 * pi) ** 2 * field.squared_frequencies
        diffuse_kernel = math.exp(fft_laplace * amount)
        return math.real(math.ifft(field.fft() * math.to_complex(diffuse_kernel)))
    else:
        data = field.data
        if isinstance(amount, Field):
            amount = amount.at(field).data
        else:
            amount = math.batch_align(amount, 0, data)
        for i in range(substeps):
            data += amount / substeps * field.laplace().data
    return field.with_data(data)


def data_bounds(field):
    assert field.has_points
    try:
        data = field.points.data
        min_vec = math.min(data, axis=tuple(range(len(data.shape) - 1)))
        max_vec = math.max(data, axis=tuple(range(len(data.shape) - 1)))
    except StaggeredSamplePoints:
        boxes = [data_bounds(c) for c in field.unstack()]
        min_vec = math.min([b.lower for b in boxes], axis=0)
        max_vec = math.max([b.upper for b in boxes], axis=0)
    return AABox(min_vec, max_vec)


def staggered_curl_2d(grid, pad_width=(1, 2)):
    assert isinstance(grid, CenteredGrid)
    kernel = math.zeros((3, 3, 1, 2))
    kernel[1, :, 0, 0] = [0, 1, -1]  # y-component: - dz/dx
    kernel[:, 1, 0, 1] = [0, -1, 1]  # x-component: dz/dy
    scalar_potential = grid.padded([pad_width, pad_width]).data
    vector_field = math.conv(scalar_potential, kernel, padding='valid')
    return StaggeredGrid(vector_field, box=grid.box)


def extrapolate(input_field, valid_mask, voxel_distance=10):
    """
    Create a signed distance field for the grid, where negative signs are fluid cells and positive signs are empty cells. The fluid surface is located at the points where the interpolated value is zero. Then extrapolate the input field into the air cells.
        :param domain: Domain that can create new Fields
        :param input_field: Field to be extrapolated
        :param valid_mask: One dimensional binary mask indicating where fluid is present
        :param voxel_distance: Optional maximal distance (in number of grid cells) where signed distance should still be calculated / how far should be extrapolated.
        :return: ext_field: a new Field with extrapolated values, s_distance: tensor containing signed distance field, depending only on the valid_mask
    """
    ext_data = input_field.data
    dx = input_field.dx
    if isinstance(input_field, StaggeredGrid):
        ext_data = input_field.staggered_tensor()
        valid_mask = math.pad(valid_mask, [[0, 0]] + [[0, 1]] * input_field.rank + [[0, 0]], "constant")

    dims = range(input_field.rank)
    # Larger than voxel_distance to be safe. It could start extrapolating velocities from outside voxel_distance into the field.
    signs = -1 * (2 * valid_mask - 1)
    s_distance = 2.0 * (voxel_distance + 1) * signs
    surface_mask = create_surface_mask(valid_mask)

    # surface_mask == 1 doesn't output a tensor, just a scalar, but >= works.
    # Initialize the voxel_distance with 0 at the surface
    # Previously initialized with -0.5*dx, i.e. the cell is completely full (center is 0.5*dx inside the fluid surface). For stability and looks this was changed to 0 * dx, i.e. the cell is only half full. This way small changes to the SDF won't directly change neighbouring empty cells to fluid cells.
    s_distance = math.where((surface_mask >= 1), -0.0 * math.ones_like(s_distance), s_distance)

    directions = np.array(list(itertools.product(
        *np.tile((-1, 0, 1), (len(dims), 1))
    )))

    # First make a move in every positive direction (StaggeredGrid velocities there are correct, we want to extrapolate these)
    if isinstance(input_field, StaggeredGrid):
        for d in directions:
            if (d <= 0).all():
                continue

            # Shift the field in direction d, compare new distances to old ones.
            d_slice = tuple(
                [(slice(1, None) if d[i] == -1 else slice(0, -1) if d[i] == 1 else slice(None)) for i in dims])

            d_field = math.pad(ext_data,
                               [[0, 0]] + [([0, 1] if d[i] == -1 else [1, 0] if d[i] == 1 else [0, 0]) for i in
                                           dims] + [[0, 0]], "symmetric")
            d_field = d_field[(slice(None),) + d_slice + (slice(None),)]

            d_dist = math.pad(s_distance,
                              [[0, 0]] + [([0, 1] if d[i] == -1 else [1, 0] if d[i] == 1 else [0, 0]) for i in dims] + [
                                  [0, 0]], "symmetric")
            d_dist = d_dist[(slice(None),) + d_slice + (slice(None),)]
            d_dist += np.sqrt((dx * d).dot(dx * d)) * signs

            if (d.dot(d) == 1) and (d >= 0).all():
                # Pure axis direction (1,0,0), (0,1,0), (0,0,1)
                updates = (math.abs(d_dist) < math.abs(s_distance)) & (surface_mask <= 0)
                updates_velocity = updates & (signs > 0)
                ext_data = math.where(
                    math.concat([(math.zeros_like(updates_velocity) if d[i] == 1 else updates_velocity) for i in dims],
                                axis=-1), d_field, ext_data)
                s_distance = math.where(updates, d_dist, s_distance)
            else:
                # Mixed axis direction (1,1,0), (1,1,-1), etc.
                continue

    for _ in range(voxel_distance):
        buffered_distance = 1.0 * s_distance  # Create a copy of current voxel_distance. This should not be necessary...
        for d in directions:
            if (d == 0).all():
                continue

            # Shift the field in direction d, compare new distances to old ones.
            d_slice = tuple(
                [(slice(1, None) if d[i] == -1 else slice(0, -1) if d[i] == 1 else slice(None)) for i in dims])

            d_field = math.pad(ext_data,
                               [[0, 0]] + [([0, 1] if d[i] == -1 else [1, 0] if d[i] == 1 else [0, 0]) for i in
                                           dims] + [[0, 0]], "symmetric")
            d_field = d_field[(slice(None),) + d_slice + (slice(None),)]

            d_dist = math.pad(s_distance, [[0, 0]] + [([0, 1] if d[i] == -1 else [1, 0] if d[i] == 1 else [0, 0]) for i in dims] + [[0, 0]], "symmetric")
            d_dist = d_dist[(slice(None),) + d_slice + (slice(None),)]
            d_dist += np.sqrt((dx * d).dot(dx * d)) * signs

            # We only want to update velocity that is outside of fluid
            updates = (math.abs(d_dist) < math.abs(buffered_distance)) & (surface_mask <= 0)
            updates_velocity = updates & (signs > 0)
            ext_data = math.where(math.concat([updates_velocity] * math.spatial_rank(ext_data), axis=-1), d_field, ext_data)
            buffered_distance = math.where(updates, d_dist, buffered_distance)

        s_distance = buffered_distance

    # Cut off inaccurate values
    distance_limit = -voxel_distance * (2 * valid_mask - 1)
    s_distance = math.where(math.abs(s_distance) < voxel_distance, s_distance, distance_limit)

    if isinstance(input_field, StaggeredGrid):
        ext_field = input_field.with_data(ext_data)
        stagger_slice = tuple([slice(0, -1) for i in dims])
        s_distance = s_distance[(slice(None),) + stagger_slice + (slice(None),)]
    else:
        ext_field = input_field.copied_with(data=ext_data)

    return ext_field, s_distance


def create_surface_mask(liquid_mask):
    """
Computes inner contours of the liquid_mask.
A cell i is flagged 1 if liquid_mask[i] = 1 and it has a non-liquid neighbour.
    :param liquid_mask: binary tensor
    :return: tensor
    """
    # When we create inner contour, we don't want the fluid-wall boundaries to show up as surface, so we should pad with symmetric edge values.
    mask = math.pad(liquid_mask, [[0, 0]] + [[1, 1]] * math.spatial_rank(liquid_mask) + [[0, 0]], "constant")
    dims = range(math.spatial_rank(mask))
    bcs = math.zeros_like(liquid_mask)

    # Move in every possible direction to assure corners are properly set.
    directions = np.array(list(itertools.product(
        *np.tile((-1, 0, 1), (len(dims), 1))
    )))

    for d in directions:
        d_slice = tuple([(slice(2, None) if d[i] == -1 else slice(0, -2) if d[i] == 1 else slice(1, -1)) for i in dims])
        center_slice = tuple([slice(1, -1) for _ in dims])

        # Create inner contour of particles
        bc_d = math.maximum(mask[(slice(None),) + d_slice + (slice(None),)],
                            mask[(slice(None),) + center_slice + (slice(None),)]) - \
            mask[(slice(None),) + d_slice + (slice(None),)]
        bcs = math.maximum(bcs, bc_d)
    return bcs



from ._config import GLOBAL_AXIS_ORDER
from ._geom_util import assert_same_rank
from ._geom import Geometry
from ._union import union  # Union is private
from ._box import AABox, BoxGenerator as _BoxGenerator
from ._sphere import Sphere

box = _BoxGenerator()  # Instantiate an AABox using the syntax box[slices]



from phi.struct.tensorop import collapsed_gather_nd

from phi.backend.dynamic_backend import DYNAMIC_BACKEND as math


def rank(tensor):
    return len(math.staticshape(tensor))


def spatial_rank(tensor):
    """ The spatial rank of a tensor is ndims - 2. """
    return math.ndims(tensor) - 2


def spatial_dimensions(obj):
    return tuple(range(1, len(math.staticshape(obj)) - 1))


def axes(obj):
    return tuple(range(len(math.staticshape(obj)) - 2))


def all_dimensions(tensor):
    return range(len(math.staticshape(tensor)))


def is_scalar(tensor):
    return math.ndims(tensor) == 0


def _get_pad_width_axes(rank, axes, val_true=(1, 1), val_false=(0, 0)):
    mid_shape = []
    for i in range(rank):
        if _contains_axis(axes, i, rank):
            mid_shape.append(val_true)
        else:
            mid_shape.append(val_false)
    return [[0, 0]] + mid_shape + [[0, 0]]


def _get_pad_width(rank, axis_widths=(1, 1)):
    return [[0, 0]] + [axis_widths] * rank + [[0, 0]]


def _dim_shifted(tensor, axis, relative_shifts, components=None, diminish_others=(0, 0), diminish_other_condition=None):
    assert len(relative_shifts) >= 2
    total_shift = max(relative_shifts) - min(relative_shifts)
    # --- Handle diminish_others ---
    if isinstance(diminish_others, tuple):
        slice_others = slice(diminish_others[0], -diminish_others[1] if diminish_others[1] != 0 else None)
    else:
        raise ValueError("Illegal diminish_others arguemnt: '%s'" % diminish_others)
    # --- Handle components ---
    if components is None:
        component_slice = slice(None)
    elif isinstance(components, int):
        component_slice = slice(components, components + 1)
    elif isinstance(components, slice):
        component_slice = components
    else:
        raise ValueError("Illegal components argument: '%s'" % components)
    # --- Slice tensor to create shifts ---
    rank = spatial_rank(tensor)
    shifted_tensors = []
    for shift in relative_shifts:
        shift_start = shift - min(relative_shifts)
        shift_end = shift_start - total_shift
        if shift_end == 0:
            shift_end = None
        slices = []
        for ax in range(rank):
            if ax == axis:
                slices.append(slice(shift_start, shift_end))
            else:
                if diminish_other_condition is None or diminish_other_condition(ax):
                    slices.append(slice_others)
                else:
                    slices.append(slice(None))
        sliced_tensor = tensor[(slice(None),) + tuple(slices) + (component_slice,)]
        shifted_tensors.append(sliced_tensor)
    return shifted_tensors


def _contains_axis(axes, axis, sp_rank):
    assert -sp_rank <= axis < sp_rank
    return (axes is None) or (axis in axes) or (axis + sp_rank in axes)


def map_for_axes(function, obj, axes, rank):
    if axes is None:
        return function(obj)
    else:
        return [(function(collapsed_gather_nd(obj, i)) if _contains_axis(axes, i, rank)
                 else collapsed_gather_nd(obj, i))
                for i in range(rank)]



from .context import unsafe
from .trait import Trait
from .structdef import definition, variable, constant, derived
from .item_condition import DATA, VARIABLES, CONSTANTS, ALL_ITEMS, ignore
from .struct import Struct, kwargs, to_dict, variables, constants, properties_dict, copy_with, isstruct, equal, VALID, INVALID

# pylint: disable-msg = redefined-builtin
from .functions import flatten, unflatten, names, map, map_item, zip, Trace, compare, print_differences, shape, staticshape, dtype, any, all



import numpy as np
import scipy
import scipy.sparse
import scipy.sparse.linalg

from phi import math
from phi.math.blas import conjugate_gradient
from phi.math.helper import _dim_shifted
from phi.physics.material import Material
from phi.struct.tensorop import collapsed_gather_nd
from .solver_api import PoissonSolver, FluidDomain


class SparseSciPy(PoissonSolver):

    def __init__(self):
        """
        The SciPy solver uses the function scipy.sparse.linalg.spsolve to determine the pressure.
        It does not support initial guesses for the pressure and does not keep track of a loop counter.
        """
        PoissonSolver.__init__(self, 'SciPy sparse solver', supported_devices=('CPU',), supports_guess=False, supports_loop_counter=False, supports_continuous_masks=True)

    def solve(self, field, domain, guess, enable_backprop):
        assert isinstance(domain, FluidDomain)
        dimensions = list(field.shape[1:-1])
        A = sparse_pressure_matrix(dimensions, domain.active_tensor(extend=1), domain.accessible_tensor(extend=1), Material.periodic(domain.domain.boundaries))

        def np_solve_p(div):
            div_vec = div.reshape([-1, A.shape[0]])
            pressure = [scipy.sparse.linalg.spsolve(A, div_vec[i, ...]) for i in range(div_vec.shape[0])]
            return np.array(pressure).reshape(div.shape).astype(np.float32)

        def np_solve_p_gradient(op, grad_in):
            return math.py_func(np_solve_p, [grad_in], np.float32, field.shape)

        pressure = math.py_func(np_solve_p, [field], np.float32, field.shape, grad=np_solve_p_gradient)
        return pressure, None


class SparseCG(PoissonSolver):

    def __init__(self, accuracy=1e-5, max_iterations=2000):
        """
        Conjugate gradient solver using sparse matrix multiplications.

        :param accuracy: the maximally allowed error for each cell, measured in terms of field values.
        :param gradient_accuracy: accuracy applied during backpropagation, number of 'same' to use forward accuracy
        :param max_iterations: integer specifying maximum conjugent gradient loop iterations or None for no limit
        :param max_gradient_iterations: maximum loop iterations during backpropagation,
            'same' uses the number from max_iterations,
            'mirror' sets the maximum to the number of iterations that were actually performed in the forward pass
        :param autodiff: If autodiff=True, use the built-in autodiff for backpropagation.
            The intermediate results of each loop iteration will be permanently stored if backpropagation is used.
            If False, replaces autodiff by a forward pressure solve in reverse accumulation backpropagation.
            This requires less memory but is only accurate if the solution is fully converged.
        """
        PoissonSolver.__init__(self, 'Sparse Conjugate Gradient', supported_devices=('CPU', 'GPU'), supports_guess=True, supports_loop_counter=True, supports_continuous_masks=True)
        assert math.is_scalar(accuracy), 'invalid accuracy: %s' % accuracy
        self.accuracy = accuracy
        self.max_iterations = max_iterations

    def solve(self, field, domain, guess, enable_backprop):
        assert isinstance(domain, FluidDomain)
        active_mask = domain.active_tensor(extend=1)
        fluid_mask = domain.accessible_tensor(extend=1)
        dimensions = math.staticshape(field)[1:-1]
        N = int(np.prod(dimensions))
        periodic = Material.periodic(domain.domain.boundaries)

        if math.choose_backend([field, active_mask, fluid_mask]).matches_name('SciPy'):
            A = sparse_pressure_matrix(dimensions, active_mask, fluid_mask, periodic)
        else:
            sidx, sorting = sparse_indices(dimensions, periodic)
            sval_data = sparse_values(dimensions, active_mask, fluid_mask, sorting, periodic)
            backend = math.choose_backend(field)
            sval_data = backend.cast(sval_data, field.dtype)
            A = backend.sparse_tensor(indices=sidx, values=sval_data, shape=[N, N])

        div_vec = math.reshape(field, [-1, int(np.prod(field.shape[1:]))])
        if guess is not None:
            guess = math.reshape(guess, [-1, int(np.prod(field.shape[1:]))])

        def apply_A(pressure): return math.matmul(A, pressure)
        result_vec, iterations = conjugate_gradient(div_vec, apply_A, guess, self.accuracy, self.max_iterations, enable_backprop)
        return math.reshape(result_vec, math.shape(field)), iterations


def sparse_pressure_matrix(dimensions, extended_active_mask, extended_fluid_mask, periodic=False):
    """
Builds a sparse matrix such that when applied to a flattened pressure channel, it calculates the laplace
of that channel, taking into account obstacles and empty cells.

    :param dimensions: valid simulation dimensions. Pressure channel should be of shape (batch size, dimensions..., 1)
    :param extended_active_mask: Binary tensor with 2 more entries in every dimension than 'dimensions'.
    :param extended_fluid_mask: Binary tensor with 2 more entries in every dimension than 'dimensions'.
    :return: SciPy sparse matrix that acts as a laplace on a flattened pressure channel given obstacles and empty cells
    """
    N = int(np.prod(dimensions))
    d = len(dimensions)
    A = scipy.sparse.lil_matrix((N, N), dtype=np.float32)
    dims = range(d)

    diagonal_entries = np.zeros(N, extended_active_mask.dtype)  # diagonal matrix entries

    gridpoints_linear = np.arange(N)
    gridpoints = np.stack(np.unravel_index(gridpoints_linear, dimensions))  # d * (N^2) array mapping from linear to spatial frames

    for dim in dims:
        lower_active, self_active, upper_active = _dim_shifted(extended_active_mask, dim, (-1, 0, 1), diminish_others=(1,1))
        lower_accessible, upper_accessible = _dim_shifted(extended_fluid_mask, dim, (-1, 1), diminish_others=(1, 1))

        stencil_upper = upper_active * self_active
        stencil_lower = lower_active * self_active
        stencil_center = - lower_accessible - upper_accessible

        diagonal_entries += math.flatten(stencil_center)

        dim_direction = math.expand_dims([1 if i == dim else 0 for i in range(d)], axis=-1)
        # --- Stencil upper cells ---
        upper_points, upper_idx = wrap_or_discard(gridpoints + dim_direction, dim, dimensions, periodic=collapsed_gather_nd(periodic, [dim, 1]))
        A[gridpoints_linear[upper_idx], upper_points] = stencil_upper.flatten()[upper_idx]
        # --- Stencil lower cells ---
        lower_points, lower_idx = wrap_or_discard(gridpoints - dim_direction, dim, dimensions, periodic=collapsed_gather_nd(periodic, [dim, 0]))
        A[gridpoints_linear[lower_idx], lower_points] = stencil_lower.flatten()[lower_idx]

    A[gridpoints_linear, gridpoints_linear] = math.minimum(diagonal_entries, -1)  # avoid 0, could lead to NaN

    return scipy.sparse.csc_matrix(A)


def sparse_indices(dimensions, periodic=False):
    N = int(np.prod(dimensions))
    d = len(dimensions)
    dims = range(d)
    gridpoints_linear = np.arange(N)
    gridpoints = np.stack(np.unravel_index(gridpoints_linear, dimensions))  # d * (N^2) array mapping from linear to spatial frames
    indices_list = [np.stack([gridpoints_linear] * 2, axis=-1)]
    for dim in dims:
        dim_direction = math.expand_dims([1 if i == dim else 0 for i in range(d)], axis=-1)
        # --- Stencil upper cells ---
        upper_points, upper_idx = wrap_or_discard(gridpoints + dim_direction, dim, dimensions, periodic=collapsed_gather_nd(periodic, [dim, 1]))
        indices_list.append(np.stack([gridpoints_linear[upper_idx], upper_points], axis=-1))
        # --- Stencil lower cells ---
        lower_points, lower_idx = wrap_or_discard(gridpoints - dim_direction, dim, dimensions, periodic=collapsed_gather_nd(periodic, [dim, 0]))
        indices_list.append(np.stack([gridpoints_linear[lower_idx], lower_points], axis=-1))
    indices = np.concatenate(indices_list, axis=0)
    # --- Sort indices ---
    sorting = np.lexsort(np.transpose(indices)[:, ::-1])
    sorted_indices = indices[sorting]
    return sorted_indices, sorting


def sparse_values(dimensions, extended_active_mask, extended_fluid_mask, sorting=None, periodic=False):
    """
    Builds a sparse matrix such that when applied to a flattened pressure channel, it calculates the laplace
    of that channel, taking into account obstacles and empty cells.

    :param dimensions: valid simulation dimensions. Pressure channel should be of shape (batch size, dimensions..., 1)
    :param extended_active_mask: Binary tensor with 2 more entries in every dimension than 'dimensions'.
    :param extended_fluid_mask: Binary tensor with 2 more entries in every dimension than 'dimensions'.
    :return: SciPy sparse matrix that acts as a laplace on a flattened pressure channel given obstacles and empty cells
    """
    N = int(np.prod(dimensions))
    d = len(dimensions)
    dims = range(d)

    values_list = []
    diagonal_entries = 0  # diagonal matrix entries

    gridpoints_linear = np.arange(N)
    gridpoints = np.stack(np.unravel_index(gridpoints_linear, dimensions))  # d * (N^2) array mapping from linear to spatial frames

    for dim in dims:
        lower_active, self_active, upper_active = _dim_shifted(extended_active_mask, dim, (-1, 0, 1), diminish_others=(1, 1))
        lower_accessible, upper_accessible = _dim_shifted(extended_fluid_mask, dim, (-1, 1), diminish_others=(1, 1))

        stencil_upper = upper_active * self_active
        stencil_lower = lower_active * self_active
        stencil_center = - lower_accessible - upper_accessible

        diagonal_entries += math.flatten(stencil_center)

        dim_direction = math.expand_dims([1 if i == dim else 0 for i in range(d)], axis=-1)
        # --- Stencil upper cells ---
        upper_points, upper_idx = wrap_or_discard(gridpoints + dim_direction, dim, dimensions, periodic=collapsed_gather_nd(periodic, [dim, 1]))
        values_list.append(math.gather(math.flatten(stencil_upper), upper_idx))
        # --- Stencil lower cells ---
        lower_points, lower_idx = wrap_or_discard(gridpoints - dim_direction, dim, dimensions, periodic=collapsed_gather_nd(periodic, [dim, 0]))
        values_list.append(math.gather(math.flatten(stencil_lower), lower_idx))

    values_list.insert(0, math.minimum(diagonal_entries, -1.))
    values = math.concat(values_list, axis=0)
    if sorting is not None:
        values = math.gather(values, sorting)
    return values


def wrap_or_discard(points, check_bounds_dim, dimensions, periodic=False):
    """
Handles points that lie outside the domain by either discarding them or wrapping them, depending on periodic.
    :param points: grid indices, typically of shape (dimensions, cell_count)
    :param check_bounds_dim: int
    :param dimensions: domain resolution
    :param periodic: if False: discard indices outside domain, if True: wrap indices outside domain
    :return:
    """
    if not periodic:
        upper_in_range_inx = np.nonzero((points[check_bounds_dim] < dimensions[check_bounds_dim]) & (points[check_bounds_dim] >= 0))[0]
        new_points = points[:, upper_in_range_inx]  # discard points outside domain
    else:
        upper_in_range_inx = slice(None)
        new_points = points % math.expand_dims(dimensions, -1)  # wrap points

    linear_points = np.ravel_multi_index(new_points, dimensions)
    return linear_points, upper_in_range_inx



import numpy as np
import phi.app.app as base_app
import six

from . import tf
from phi.app.app import EditableFloat, EditableInt, EditableValue
from phi.data.dataset import Dataset
from phi.data.reader import BatchReader
from phi.physics.field import Field, StaggeredGrid
from phi.tf.data import Dataset as TFDataset

from . import TF_BACKEND
from .session import Session
from .world import tf_bake_graph
from .. import math
from ..app import EditableBool


class App(base_app.App):

    def __init__(self, *args, **kwargs):
        base_app.App.__init__(self, *args, **kwargs)
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        self.session = Session(self.scene, session=tf.Session(config=config))
        self.scalars = []
        self.scalar_names = []
        self.editable_placeholders = {}  # placeholder -> attribute name
        self.auto_bake = True
        self.add_trait('tensorflow')

    def prepare(self):
        if self.prepared:
            return
        base_app.App.prepare(self)
        if self.auto_bake:
            self.info("Baking static TensorFlow graph (disable by setting 'app.auto_bake = False'")
            tf_bake_graph(self.world, self.session)
        self.info('Initializing variables')
        self.session.initialize_variables()
        return self

    def add_scalar(self, name, node):
        assert TF_BACKEND.is_tensor(node), 'add_scalar requires a TensorFlow tensor but got %s' % node
        self.scalar_names.append(name)
        self.scalars.append(node)

    def editable_float(self, name, initial_value, minmax=None, log_scale=None):
        val = EditableFloat(name, initial_value, minmax, None, log_scale)
        self.set_editable_value(name, val)
        placeholder = tf.placeholder(TF_BACKEND.precision_dtype, (), name.lower().replace(' ', '_'))
        self.add_scalar(name, placeholder)
        self.editable_placeholders[placeholder] = name
        return placeholder

    def editable_int(self, name, initial_value, minmax=None):
        val = EditableInt(name, initial_value, minmax, None)
        self.set_editable_value(name, val)
        placeholder = tf.placeholder(tf.int32, (), name.lower().replace(' ', '_'))
        self.add_scalar(name, placeholder)
        self.editable_placeholders[placeholder] = name
        return placeholder

    def editable_bool(self, name, initial_value, add_as_scalar=False):
        val = EditableBool(name, initial_value)
        self.set_editable_value(name, val)
        placeholder = tf.placeholder(tf.bool, (), name.lower().replace(' ', '_'))
        if add_as_scalar:
            self.add_scalar(name, math.to_float(placeholder))
        self.editable_placeholders[placeholder] = name
        return placeholder

    def editable_values_dict(self):
        return {placeholder: self.get_editable_value(name) for placeholder, name in self.editable_placeholders.items()}

    def get_editable_value(self, name):
        value = getattr(self, '_ed_val_' + name.lower())
        if isinstance(value, EditableValue):
            return value.initial_value
        else:
            return value

    def set_editable_value(self, name, value):
        setattr(self, '_ed_val_' + name.lower(), value)


def EVERY_EPOCH(tfapp): return tfapp.steps % tfapp.epoch_size == 0


class LearningApp(App):

    def __init__(self, name='TensorFlow application', subtitle='',
                 learning_rate=1e-3,
                 training_batch_size=4,
                 validation_batch_size=16,
                 model_scope_name='model',
                 base_dir='~/phi/model/',
                 stride=None,
                 epoch_size=None,
                 force_custom_stride=False,
                 log_scalars=EVERY_EPOCH,
                 **kwargs):
        App.__init__(self, name=name, subtitle=subtitle, base_dir=base_dir, **kwargs)
        self.add_trait('model')

        # --- Model ---
        self.model_scope_name = model_scope_name
        self.auto_bake = False
        self.scalar_values = {}
        self.scalar_values_validation = {}
        self.learning_rate = self.editable_float('Learning_Rate', learning_rate)
        self.all_optimizers = []
        # --- Data ---
        self.training = tf.placeholder(tf.bool, (), 'training')
        self.training_batch_size = training_batch_size
        self.validation_batch_size = validation_batch_size
        self._placeholder_struct = None  # Data-placeholders or Iterator placeholder
        self._training_set = None
        self._validation_set = None
        self._pipeline = None
        self.set_data(None, None)
        assert stride is None or epoch_size is None
        self.epoch_size = epoch_size if epoch_size is not None else stride
        assert isinstance(log_scalars, bool) or callable(log_scalars)
        self.log_scalars = log_scalars

    def prepare(self):
        if self.prepared:
            return

        scalars = [tf.summary.scalar(self.scalar_names[i], self.scalars[i]) for i in range(len(self.scalars))]
        self.merged_scalars = tf.summary.merge(scalars)

        App.prepare(self)  # initializes global variables

        model_parameter_count = 0
        for var in tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=self.model_scope_name):
            if not 'Adam' in var.name:
                model_parameter_count += int(np.prod(var.get_shape().as_list()))
                # if 'conv' in var.name and 'kernel' in var.name:
                #     tf.summary.image(var.name, var)
        self.add_custom_property('parameter_count', model_parameter_count)
        self.info('Model variables contain %d total parameters.' % model_parameter_count)
        # --- Use world.batch_size? ---
        if self.world.batch_size is not None:
            self.training_batch_size = self.world.batch_size
            self.validation_batch_size = self.world.batch_size
        # --- Epoch size ---
        if self.epoch_size is None:
            if self._train_reader is not None:
                self.epoch_size = len(self._train_reader.all_batches(batch_size=self.training_batch_size))
            else:
                self.epoch_size = 1
        self.sequence_stride = self.epoch_size
        # --- Validate ---
        self.validation_step()
        return self

    def set_learning_rate(self, learning_rate):
        self.set_editable_value('Learning_Rate', learning_rate)

    def set_data(self, dict, train=None, val=None):
        """
Specify what data to use for training and validation.

The content of `dict` determines the data pipeline that is used.
  - 'placeholder' pipline: the static TensorFlow graph uses placeholders as input. `dict` maps from placeholders to file names or Stream instances. Placeholders and corresponding streams may be placed inside structs.
  - 'dataset_handle' pipeline: Use TensorFlow data pipeline. `dict` contains 'iterator_handle' and related properties as returned by `build_graph_input(...)[1]`.

Regardless of pipeline, the recommended way to obtain `dict` is through `build_graph_input(...)[1]`.

        :param dict: pipeline-dependent dict
        :type dict: dict
        :param train: (optional) Dataset used for training
        :type train: Dataset
        :param val: (optional) Dataset used for validation
        :type val: Dataset
        """
        assert isinstance(train, Dataset) or train is None
        assert isinstance(val, Dataset) or val is None
        if train is not None or val is not None:
            assert dict is not None
        if train is not None and val is not None:
            self.value_view_training_data = False
        self._training_set = train
        self._validation_set = val
        if dict is not None and 'iterator_handle' in dict:
            self._init_tf_pipeline(**dict)
        else:
            self._init_numpy_iterators(dict)

    def _init_numpy_iterators(self, dict):
        self._pipeline = 'placeholder'
        self._placeholder_struct = []
        self._channel_struct = []
        if dict is not None:
            for key, value in dict.items():
                self._placeholder_struct.append(key)
                self._channel_struct.append(value)
        self._channel_struct = tuple(self._channel_struct)
        self._placeholder_struct = tuple(self._placeholder_struct)
        # Train
        if self._training_set is not None:
            self._train_reader = BatchReader(self._training_set, self._channel_struct)
            self._train_iterator = self._train_reader.all_batches(batch_size=self.world.batch_size or self.training_batch_size, loop=True)
        else:
            self._train_reader = None
            self._train_iterator = None
        # Val
        if self._validation_set is not None:
            self._val_reader = BatchReader(self._validation_set, self._channel_struct)
        else:
            self._val_reader = None

    def _init_tf_pipeline(self, iterator_handle, names, shapes, dtypes, frames):
        self._placeholder_struct = iterator_handle
        self._pipeline = 'dataset_handle'
        if self._training_set is not None:
            assert isinstance(self._training_set, TFDataset)
            if self._training_set.name is None:
                self._training_set.name = 'train'
            self._training_set.setup(names, shapes, dtypes, batch_size=self.training_batch_size, frames=frames)
            self._training_set.reset_iterator(self.session)
        if self._validation_set is not None:
            assert isinstance(self._validation_set, TFDataset)
            if self._validation_set.name is None:
                self._validation_set.name = 'validation'
            self._validation_set.setup(names, shapes, dtypes, batch_size=self.validation_batch_size, frames=frames)

    def add_objective(self, loss, name='Loss', optimizer=None, reg=None, vars=None):
        assert len(loss.shape) <= 1, 'Loss function must be a scalar'
        if not optimizer:
            optimizer = tf.train.AdamOptimizer(self.learning_rate)

        if reg is not None:
            self.add_scalar(name + '_reg_unscaled', reg)
            reg_scale = self.editable_float(name + '_reg_scale', 1.0)
            optim_function = loss + reg * reg_scale
        else:
            optim_function = loss

        if isinstance(vars, six.string_types):
            vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=vars)

        node = optimizer.minimize(optim_function, var_list=vars)

        self.add_scalar(name, loss)
        self.all_optimizers.append(node)
        return node

    def step(self):
        optimized = self.optimization_step(self.all_optimizers)
        if not optimized:
            self.steps -= 1
        if self._pipeline == 'placeholder' and self.steps % self.epoch_size == 0:
            self.validation_step(create_checkpoint=True)
        return self

    def optimization_step(self, optim_nodes, log_loss=None, summary_key='train'):
        try:
            optim_nodes = list(optim_nodes)
        except:
            optim_nodes = [optim_nodes]
        if self._pipeline == 'placeholder':
            batch = next(self._train_iterator) if self._train_iterator is not None else None
        elif self._pipeline == 'dataset_handle':
            batch = self._training_set.iterator_handle
        else:
            raise NotImplementedError('Pipeline %s' % self._pipeline)
        feed_dict = self._feed_dict(batch, True)
        try:
            scalar_values = self.session.run(optim_nodes + self.scalars, feed_dict, summary_key=summary_key, merged_summary=self.merged_scalars, time=self.steps)[len(optim_nodes):]
        except tf.errors.OutOfRangeError as error:
            if self._pipeline != 'dataset_handle':
                raise error
            self.on_training_set_end()
            return False
        self.scalar_values = {name: value for name, value in zip(self.scalar_names, scalar_values)}
        if log_loss is None:
            log_loss = self.log_scalars
        if callable(log_loss):
            log_loss = log_loss(self)
        assert isinstance(log_loss, bool)
        if log_loss:
            self.info('Optimization (%06d): ' % self.steps + ', '.join([self.scalar_names[i] + ': ' + str(scalar_values[i]) for i in range(len(self.scalars))]))
        return True

    def on_training_set_end(self):
        self.validation_step(create_checkpoint=True)
        self._training_set.reset_iterator(self.session)

    def validation_step(self, create_checkpoint=False):
        if self._validation_set is None:
            return
        if self._pipeline == 'placeholder':
            batch = self._val_reader[0:self.validation_batch_size]
        elif self._pipeline == 'dataset_handle':
            batch = self._validation_set.get_reset_handle(self.session)
        else:
            raise NotImplementedError('Pipeline %s' % self._pipeline)
        feed_dict = self._feed_dict(batch, False)
        # ToDo iterate over complete valiadtion set and average the results, e.g. with tf.contrib.metrics.streaming_mean - https://stackoverflow.com/questions/40788785/how-to-average-summaries-over-multiple-batches
        scalar_values = self.session.run(self.scalars, feed_dict, summary_key='val', merged_summary=self.merged_scalars, time=self.steps)
        self.scalar_values_validation = {name: value for name, value in zip(self.scalar_names, scalar_values)}
        if create_checkpoint:
            self.save_model()
        self.info('Validation (%06d): ' % self.steps + ', '.join([self.scalar_names[i] + ': ' + str(scalar_values[i]) for i in range(len(self.scalars))]))

    def base_feed_dict(self):
        return {}

    def _feed_dict(self, batch, training):
        """
Assemble a complete feed dict for graph execution.
        :param batch: (optional)
          'placeholder' pipeline: struct of Numpy arrays matching self._placeholder_struct
          'dataset_handle' pipeline: iterator_handle string value
        :param training: bool
        :return: dict that can be passed to session.run()
        """
        feed_dict = self.base_feed_dict()
        feed_dict.update(self.editable_values_dict())
        feed_dict[self.training] = training
        if batch is not None:
            feed_dict[self._placeholder_struct] = batch
        return feed_dict

    @property
    def view_reader(self):
        if self._val_reader is None and self._train_reader is None:
            return None
        if self._val_reader is None:
            return self._train_reader
        return self._train_reader if self.value_view_training_data else self._val_reader

    def _view_dataset(self):
        if self._validation_set is None and self._training_set is None:
            return None
        if self._validation_set is None:
            return self._training_set
        if self.value_view_training_data:
            return self._training_set
        else:
            return self._validation_set

    def view(self, tasks):
        if tasks is None:
            return None
        if self._pipeline == 'placeholder':
            batch = self.view_reader[0:self.validation_batch_size] if self.view_reader is not None else None
        elif self._pipeline == 'dataset_handle':
            batch = self._view_dataset().get_reset_handle(self.session)
        else:
            raise NotImplementedError('Pipeline %s' % self._pipeline)
        feed_dict = self._feed_dict(batch, False)
        return self.session.run(tasks, feed_dict)

    @property
    def viewed_batch(self):
        assert self.view_reader is not None, 'There is no data to view.'
        return self.view_reader[0:self.validation_batch_size]

    def view_batch(self, get_attribute):
        batch = self.view_reader[0:self.validation_batch_size]
        return get_attribute(batch)

    def save_model(self):
        dir = self.scene.subpath('checkpoint_%08d' % self.steps)
        self.session.save(dir)
        return dir

    def action_save_model(self):
        self.save_model()

    def load_model(self, checkpoint_dir):
        self.session.restore(checkpoint_dir, scope=self.model_scope_name)

    def model_scope(self):
        return tf.variable_scope(self.model_scope_name)

    def add_field(self, name, field):
        """

        :param name: channel name
        :param field: Tensor, string (database fieldname) or function
        """
        if is_tensorflow_field(field):
            App.add_field(self, name, lambda: self.view(field))
        else:
            App.add_field(self, name, field)


TFApp = LearningApp


def is_tensorflow_field(obj):
    if TF_BACKEND.is_tensor(obj):
        return True
    if isinstance(obj, StaggeredGrid):
        return np.any([is_tensorflow_field(grid) for grid in obj.data])
    if isinstance(obj, Field):
        return TF_BACKEND.is_tensor(obj.data)
    return False



import warnings

from . import tf
from phi import struct, math
from phi.data.fluidformat import _transform_for_writing, _writing_staticshape, read_zipped_array, _slugify_filename
from phi.math import is_static_shape
from phi.physics.world import StateProxy
from phi.struct.context import _unsafe
from phi.data import SceneSource, Dataset as BaseDataset

from .util import placeholder, dataset_handle
from ..data.stream import consecutive_frames, FrameSelect


def build_graph_input(obj, input_type='placeholder', frames=None, names=None):
    """
Create placeholders for tensors in the supplied state.
    :param names: structure compatible with `obj` holding the associated file names
    :param obj: struct or StateProxy
    :param input_type: 'placeholder' or 'dataset_handle'
    :param frames: Number of input frames. If not None, returns a list of input structs.
    :return:
      1. Valid state containing or derived from created placeholders or dataset handle
      2. dict mapping from placeholders to their default names (using struct.names)
    """
    if isinstance(obj, StateProxy):
        obj = obj.state
    assert struct.isstruct(obj)
    # --- Shapes and names ---
    writable_obj = _transform_for_writing(obj)
    shape = _writing_staticshape(obj)
    if names is None:
        names = struct.names(writable_obj)
        names = struct.map(_slugify_filename, names, content_type=struct.names)
    if input_type == 'placeholder':
        if frames is not None:
            with _unsafe():
                placeholders = [placeholder(shape, basename='Placeholder_frame_%d' % i) for i in range(frames)]
            graph_in = [struct.map(lambda x: x, p) for p in placeholders]  # validates fields, splits staggered tensors
            streams = struct.map(lambda x: consecutive_frames(x, frames), names, content_type='streams')
            load_dict = {placeholders[i]: struct.map(lambda x: x[i], streams, leaf_condition=lambda x: isinstance(x, tuple) and isinstance(x[0], FrameSelect), content_type='streams') for i in range(frames)}
            return graph_in, load_dict
        else:
            with _unsafe():
                placeholders = placeholder(shape)
            graph_in = struct.map(lambda x: x, placeholders)  # validates fields, splits staggered tensors
            return graph_in, {placeholders: names}
    elif input_type == 'dataset_handle':
        with _unsafe():
            dtypes = struct.dtype(writable_obj)
            dataset_nodes, iterator_handle = dataset_handle(shape, dtypes, frames=frames)
        graph_in = struct.map(lambda x: x, dataset_nodes)  # validates fields, splits staggered tensors
        shapes = struct.flatten(struct.staticshape(dataset_nodes), leaf_condition=is_static_shape)
        dtypes = struct.flatten(struct.dtype(dataset_nodes))
        return graph_in, {'names': struct.flatten(names), 'iterator_handle': iterator_handle, 'shapes': shapes, 'dtypes': dtypes, 'frames': frames}
    else:
        raise ValueError(input_type)


def load_state(state):
    warnings.warn("load_state() is deprecated, use build_graph_input() instead.")
    return build_graph_input(state)


def create_dataset(scene_sources, names, shapes, dtypes, batch_size, frames=None, shuffle=False, inner_frame_stride=1, outer_frame_stride=1, prefetch=2):
    count = 0
    datasets = []
    for source in scene_sources:
        scene = source.scene
        nested_file_list = list(scene.data_paths(source.frames(), field_names=names))
        count += _example_count(len(nested_file_list), frames, inner_frame_stride, outer_frame_stride)
        dataset = tf.data.Dataset.from_tensor_slices(nested_file_list)
        dataset = dataset.map(lambda *items: tuple(tf.py_func(_read_npy_files, items, dtypes)))
        if frames is not None:
            dataset = stacked_window(dataset, frames, outer_stride=outer_frame_stride, inner_stride=inner_frame_stride)
        datasets.append(dataset)
    dataset = concat_datasets(datasets)
    if shuffle:
        dataset = dataset.shuffle(count)
    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(prefetch)
    return dataset


def _read_npy_files(items):
    data = [read_zipped_array(item.decode())[0,...] for item in items]
    data = [math.to_float(array) for array in data]
    return data


def stacked_window(dataset, size, outer_stride=1, inner_stride=1):
    """
Combines lists of input elements into windows by adding a window dimension to the dataset.
All windows have the same number of elements.
    :param dataset: TensorFlow Dataset
    :param size: number of elements in each window
    :param outer_stride: skip windows; makes the outer dataset smaller
    :param inner_stride: element-stride inside each window
    :return: Dataset
    """
    dataset = dataset.window(size, shift=outer_stride, stride=inner_stride, drop_remainder=True)
    dataset = dataset.flat_map(lambda x: x)  # convert VariantDataset (window) to batch dimension
    dataset = dataset.batch(size)
    return dataset


def concat_datasets(datasets):
    """ Creates a TensorFlow Dataset from the given ordered list of TensorFlow Datasets """
    # n-1 concatenations will lead to overflow (RecursionError) if list is too long
    # Instead, concatenate them hierarchically
    assert len(datasets) > 0
    if len(datasets) == 1:
        return datasets[0]
    else:
        center = len(datasets) // 2
        dataset_1 = concat_datasets(datasets[:center])
        dataset_2 = concat_datasets(datasets[center:])
        return tf.data.Dataset.concatenate(dataset_1, dataset_2)


def _example_count(length, frames, inner_stride, outer_stride):
    if frames is None:
        return length
    return (frames - (frames * inner_stride - 1)) // outer_stride


class Dataset(BaseDataset):
    """
Extends phi.data.Datset by TensorFlow data pipeline functions.
    """

    def __init__(self, name, sources):
        BaseDataset.__init__(self, name, sources)
        self.shuffled = False
        self.prefetch_value = 1
        self.inner_frame_stride = 1
        self.outer_frame_stride = 1
        self.batch_size = None
        self.tf_dataset = None
        self.iterator = None
        self.iterator_handle = None

    @staticmethod
    def load(directory, indices=None, name=None, max_scenes=None, assume_same_frames=True, assume_same_shapes=True, frames=None):
        base = BaseDataset.load(directory, indices=indices, name=name, max_scenes=max_scenes, assume_same_frames=assume_same_frames, assume_same_shapes=assume_same_shapes, frames=frames)
        return Dataset(base.name, base.sources)

    def shuffle(self):
        assert self.tf_dataset is None
        self.shuffled = True
        return self

    def prefetch(self, prefetch):
        assert self.tf_dataset is None
        self.prefetch_value = prefetch
        return self

    def batch(self, batch_size):
        assert self.tf_dataset is None
        self.batch_size = batch_size
        return self

    def setup(self, names, shapes, dtypes, batch_size=None, frames=None):
        for source in self.sources:
            assert isinstance(source, SceneSource)
        batch_size = batch_size if batch_size is not None else self.batch_size
        batch_size = 1 if batch_size is None else batch_size
        self.tf_dataset = create_dataset(self.sources, names=names, shapes=shapes, dtypes=dtypes, batch_size=batch_size, frames=frames, shuffle=self.shuffled, inner_frame_stride=self.inner_frame_stride, outer_frame_stride=self.outer_frame_stride, prefetch=self.prefetch_value)
        self.iterator = self.tf_dataset.make_initializable_iterator()

    def reset_iterator(self, session):
        if self.iterator_handle is None:
            self.iterator_handle = session.run(self.iterator.string_handle())
        session.run(self.iterator.initializer)

    def get_reset_handle(self, session):
        self.reset_iterator(session)
        return self.iterator_handle



import numpy as np


def _is_leaf(tensor_like, leaf_condition):
    if not isinstance(tensor_like, (tuple, list, np.ndarray)):
        return True
    if leaf_condition is not None and leaf_condition(tensor_like):
        return True
    return False


def collapse(tensor_like, leaf_condition=None):
    if _is_leaf(tensor_like, leaf_condition):
        return tensor_like
    collapsed_elements = tuple([collapse(element, leaf_condition) for element in tensor_like])
    first = collapsed_elements[0]
    for element in collapsed_elements[1:]:
        if element != first:
            return collapsed_elements
    return first


def collapsed_gather_nd(collapsed, nd_index, leaf_condition=None):
    if isinstance(collapsed, (tuple, list, np.ndarray)):
        if leaf_condition is not None and leaf_condition(collapsed):
            return collapsed
        # collapsed = np.array(collapsed)
        if len(nd_index) == 1:
            return collapsed[nd_index[0]]
        else:
            return collapsed_gather_nd(collapsed[nd_index[0]], nd_index[1:])
    else:
        return collapsed


def expand(collapsed, shape):
    if len(shape) == 0:
        return collapsed
    if isinstance(collapsed, (tuple, list, np.ndarray)):
        if len(collapsed) == shape[0]:
            return [expand(item, shape[1:]) for item in collapsed]
        elif len(collapsed) == 1:
            item = expand(collapsed[0], shape[1:])
            return [item] * shape[0]
        else:
            raise ValueError('Cannot match shape: requested %d but actual %d' % (shape[0], len(collapsed)))
    else:
        return [expand(collapsed, shape[1:])] * shape[0]


class CollapsedTensor(object):

    def __init__(self, collapsed, leaf_condition=None, shape=None):
        self.collapsed = collapsed
        self.leaf_condition = leaf_condition
        self.shape = shape

    def __getitem__(self, item):
        return collapsed_gather_nd(self.collapsed, item, self.leaf_condition)

    def expand(self, shape=None):
        shape = self.shape if shape is None else shape
        assert shape is not None
        return expand(self.collapsed, shape)



import numpy as np

from phi import struct, math
from ._geom_util import assert_same_rank
from ._geom import Geometry
from ._transform import rotate


@struct.definition()
class AbstractBox(Geometry):

    @property
    def size(self):
        raise NotImplementedError(self)

    @property
    def half_size(self):
        raise NotImplementedError(self)

    @property
    def lower(self):
        raise NotImplementedError(self)

    @property
    def upper(self):
        raise NotImplementedError(self)

    def bounding_radius(self):
        return math.max(self.size, axis=-1, keepdims=True) * 1.414214

    def bounding_half_extent(self):
        return self.size * 0.5

    def global_to_local(self, global_position):
        size, lower = math.batch_align([self.size, self.lower], 1, global_position)
        return (global_position - lower) / size

    def local_to_global(self, local_position):
        size, lower = math.batch_align([self.size, self.lower], 1, local_position)
        return local_position * size + lower

    def lies_inside(self, location):
        lower, upper = math.batch_align([self.lower, self.upper], 1, location)
        bool_inside = (location >= lower) & (location <= upper)
        return math.all(bool_inside, axis=-1, keepdims=True)

    def approximate_signed_distance(self, location):
        """
Computes the signed L-infinity norm (manhattan distance) from the location to the nearest side of the box.
For an outside location `l` with the closest surface point `s`, the distance is `max(abs(l - s))`.
For inside locations it is `-max(abs(l - s))`.
        :param location: float tensor of shape (batch_size, ..., rank)
        :return: float tensor of shape (*location.shape[:-1], 1).
        """
        lower, upper = math.batch_align([self.lower, self.upper], 1, location)
        center = 0.5 * (lower + upper)
        extent = upper - lower
        distance = math.abs(location - center) - extent * 0.5
        return math.max(distance, axis=-1, keepdims=True)

    def get_lower(self, axis):
        return self._get(self.lower, axis)

    def get_upper(self, axis):
        return self._get(self.upper, axis)

    @staticmethod
    def _get(vector, axis):
        if vector.shape[-1] == 1:
            return vector[...,0]
        else:
            return vector[...,axis]

    def corner_representation(self):
        return AABox(self.lower, self.upper)

    def center_representation(self):
        return Cuboid(self.center, self.half_size)

    def contains(self, other):
        if isinstance(other, AbstractBox):
            return np.all(other.lower >= self.lower) and np.all(other.upper <= self.upper)
        else:
            raise NotImplementedError()

    def rotated(self, angle):
        return rotate(self, angle)


@struct.definition(traits=[math.BATCHED])
class AABox(AbstractBox):
    """
    Axis-aligned box, defined by lower and upper corner.
    AABoxes can be created using the shorthand notation box[slices], (e.g. box[:,0:1] to create an inifinite-height box from x=0 to x=1).
    """

    def __init__(self, lower, upper, **kwargs):
        AbstractBox.__init__(self, **struct.kwargs(locals()))

    @struct.constant(min_rank=1)
    def lower(self, lower):
        return math.to_float(lower)

    @struct.constant(min_rank=1)
    def upper(self, upper):
        return math.to_float(upper)

    @property
    def rank(self):
        if math.ndims(self.size) > 0:
            return math.staticshape(self.size)[-1]
        else:
            return None

    @struct.derived()
    def size(self):
        return self.upper - self.lower

    @struct.derived()
    def center(self):
        return 0.5 * (self.lower + self.upper)

    @struct.derived()
    def half_size(self):
        return self.size * 0.5

    def without_axis(self, axis):
        lower = []
        upper = []
        for ax in range(self.rank):
            if ax != axis:
                lower.append(self.get_lower(ax))
                upper.append(self.get_upper(ax))
        return self.copied_with(lower=lower, upper=upper)

    def shifted(self, delta):
        return self.copied_with(lower=self.lower + delta, upper=self.upper + delta)

    def __repr__(self):
        if self.is_valid:
            return '%s at (%s)' % ('x'.join([str(x) for x in self.size]), ','.join([str(x) for x in self.lower]))
        else:
            return struct.Struct.__repr__(self)

    @staticmethod
    def to_box(value, resolution_hint=None):
        if value is None:
            assert resolution_hint is not None
            result = AABox(0, resolution_hint)
        elif isinstance(value, AABox):
            result = value
        elif isinstance(value, int):
            if resolution_hint is None:
                result = AABox(0, value)
            else:
                size = [value] * (1 if math.ndims(resolution_hint) == 0 else len(resolution_hint))
                result = AABox(0, size)
        elif isinstance(value, (tuple, list)):
            result = AABox(0, box)
        else:
            raise ValueError("Box extent not understood: '%s'" % value)
        if resolution_hint is not None:
            assert_same_rank(len(resolution_hint), result, 'AABox rank does not match resolution.')
        return result


@struct.definition(traits=[math.BATCHED])
class Cuboid(AbstractBox):

    def __init__(self, center, half_size, **kwargs):
        AbstractBox.__init__(self, **struct.kwargs(locals()))

    @struct.constant(min_rank=1)
    def center(self, center):
        return math.to_float(center)

    @struct.constant(min_rank=1)
    def half_size(self, half_size):
        return math.to_float(half_size)

    @property
    def rank(self):
        if math.ndims(self.upper) > 0:
            return math.staticshape(self.upper)[-1]
        else:
            return None

    @struct.derived()
    def size(self):
        return 2 * self.half_size

    @struct.derived()
    def lower(self):
        return self.center - self.half_size

    @struct.derived()
    def upper(self):
        return self.center + self.half_size

    def shifted(self, delta):
        return self.copied_with(center=self.center + delta)


class BoxGenerator(object):

    def __getitem__(self, item):
        if not isinstance(item, (tuple, list)):
            item = [item]
        lower = []
        upper = []
        for dim in item:
            assert isinstance(dim, slice)
            assert dim.step is None or dim.step == 1, "Box: step must be 1 but is %s" % dim.step
            lower.append(dim.start if dim.start is not None else -np.inf)
            upper.append(dim.stop if dim.stop is not None else np.inf)
        return AABox(lower, upper)

    def __call__(self, *args, **kwargs):
        if len(args) == 2 and len(kwargs) == 0:
            lower, upper = args
            return AABox(lower, upper)
        elif len(args) == 1 and 'size' in kwargs:
            center, = args
            size = kwargs['size']
            return Cuboid(center, 0.5 * math.to_float(size))
        elif 'size' in kwargs and 'center' in kwargs:
            center, size = kwargs['center'], kwargs['size']
            return Cuboid(center, 0.5 * math.to_float(size))
        elif len(args) == 1 and len(kwargs) == 0:
            return AABox(0, args[0])
        else:
            raise ValueError('Cannot create box from args=%s, kwargs=%s' % (args, kwargs))


def bounding_box(geometry):
    center = geometry.center
    extent = geometry.bounding_half_extent()
    return AABox(lower=center - extent, upper=center + extent)



import six
import numpy as np

from .source import DataSource


class DataStream(object):

    def shape(self, datasource):
        raise NotImplementedError(self)

    def size(self, datasource, lookup=False):
        raise NotImplementedError(self)

    def get(self, datasource, indices):
        raise NotImplementedError(self)

    def frames(self, datasource):
        raise NotImplementedError(self)

    def __add__(self, other):
        if isinstance(other, DataStream):
            return ElementwiseOperationStream([self, other], lambda a, b: a + b)
        else:
            return ElementwiseOperationStream([self], lambda a: a + other)

    def __sub__(self, other):
        if isinstance(other, DataStream):
            return ElementwiseOperationStream([self, other], lambda a, b: a - b)
        else:
            return ElementwiseOperationStream([self], lambda a: a - other)

    def __mul__(self, other):
        if isinstance(other, DataStream):
            return ElementwiseOperationStream([self, other], lambda a, b: a * b)
        else:
            return ElementwiseOperationStream([self], lambda a: a * other)

    def __div__(self, other):
        if isinstance(other, DataStream):
            return ElementwiseOperationStream([self, other], lambda a, b: a / b)
        else:
            return ElementwiseOperationStream([self], lambda a: a / other)


class SourceStream(DataStream):

    def __init__(self, name):
        assert isinstance(name, six.string_types)
        self._name = name

    def shape(self, datasource):
        return datasource.shape(self._name)

    def size(self, datasource, lookup=False):
        return datasource.size(lookup=lookup)

    def get(self, datasource, indices):
        frames = datasource.frames()
        frames = [frames[i] for i in indices]
        return datasource.get(self._name, frames)

    def frames(self, datasource):
        return datasource.frames()

    def __repr__(self):
        return self._name


class _SourceFrame(DataStream):

    def shape(self, datasource):
        return (1, )

    def size(self, datasource, lookup=False):
        return datasource.size(lookup=lookup)

    def get(self, datasource, indices):
        frames = datasource.frames()
        return np.expand_dims([frames[i] for i in indices], -1)

    def frames(self, datasource):
        return datasource.frames()


FRAME = _SourceFrame()


class _SourceScene(DataStream):

    def shape(self, datasource):
        return (1, )

    def size(self, datasource, lookup=False):
        return datasource.size(lookup=lookup)

    def get(self, datasource, indices):
        return np.expand_dims([datasource.scene] * len(indices), -1)

    def frames(self, datasource):
        return datasource.frames()


SCENE = _SourceScene()


class _Source(DataStream):

    def shape(self, datasource):
        return (1, )

    def size(self, datasource, lookup=False):
        return datasource.size(lookup=lookup)

    def get(self, datasource, indices):
        return np.expand_dims([datasource] * len(indices), -1)

    def frames(self, datasource):
        return datasource.frames()


SOURCE = _Source()


class DerivedStream(DataStream):

    def __init__(self, input_streams):
        self.inputs = [c if isinstance(c, DataStream) else SourceStream(c) for c in input_streams]

    def __repr__(self):
        return "%s(%s)" % (type(self), self.inputs)


class ElementwiseOperationStream(DerivedStream):

    def __init__(self, input_streams, function):
        DerivedStream.__init__(self, input_streams)
        self.function = function

    def shape(self, datasource):
        return self.inputs[0].shape(datasource)

    def size(self, datasource, lookup=False):
        return self.inputs[0].size(datasource, lookup=lookup)

    def frames(self, datasource):
        return self.inputs[0].frames(datasource)

    def get(self, datasource, indices):
        for index in indices:
            input_values = [np.concatenate(list(i.get(datasource, [index]))) for i in self.inputs]
            result = self.function(*input_values)
            yield result


class FrameSelect(DerivedStream):

    def __init__(self, selector, stream):
        """
        Selects specific frames from the input.

        :param selector: Either a frame index, list of frame frames or a selection function mapping a list of all frames to a list of selected frames
        :param stream: input stream
        """
        DerivedStream.__init__(self, [stream])
        self.stream = self.inputs[0]
        if callable(selector):
            self.selection_function = selector
        else:
            if isinstance(selector, int):
                self.selection_function = lambda frames: [selector]
            else:
                self.selection_function = lambda frames: selector

    def shape(self, datasource):
        return self.stream.shape(datasource)

    def get(self, datasource, indices):
        input_frames = self.stream.frames(datasource)
        frames = self.selection_function(input_frames)
        if isinstance(frames, int):
            frames = [frames]
        try:
            selected_frames = [frames[i] for i in indices]
        except:
            raise ValueError("BatchSelect: selection function must return a list of integers that is large enough, but got %s for frames %s" % (frames, datasource.frames()))
        return self.stream.get(datasource, selected_frames)

    def size(self, datasource, lookup=False):
        if not lookup:
            size = datasource.size()
            if size is None:
                return None
        input_frames = self.stream.frames(datasource)
        frames = self.selection_function(input_frames)
        return 1 if isinstance(frames, int) else len(frames)

    def frames(self, datasource):
        input_frames = self.stream.frames(datasource)
        frames = self.selection_function(input_frames)
        return [frames] if isinstance(frames, int) else frames


class MantaScalar(DerivedStream):

    def __init__(self, stream):
        """
        Removes one layer of cells on the positive sides for scalar streams.
        This can be used to load mantaflow fluid sim scenes, for which the staggered velocity
        will be loaded unmodified, while the scalar grids are cropped to match the size
        of the phiflow arrays.
        """
        DerivedStream.__init__(self, [stream])
        self.stream = self.inputs[0]

    def shape(self, datasource):
        return self.stream.shape(datasource)

    def get(self, datasource, indices):
        a = self.stream.get(datasource, indices)
        c = []
        for b in a:
            b = b[..., 0:b.shape[1] - 1, 0:b.shape[2] - 1, :]  # crop 1 layer
            c.append(b)
        a = np.asarray(c)
        return a

    def size(self, datasource, lookup=False):
        return self.stream.size(datasource, lookup)

    def frames(self, datasource):
        return self.stream.frames(datasource)


def consecutive_frames(channel, n):
    """
Constructs n DataChannels to load a consequtive sequence of frames.
For each sequence, the i-th DataChannel loads the i-th frame in that sequence.
The number of sequences depends on the number of frames contained in each source.

The frames of each source are assumed to be ordered and without frame gaps.
Every DataSource accessed through these channels must contain at least n frames.
    :param channel: Source channel holding the data
    :param n: Number of frames to load
    :return: tuple containing n DataChannels
    """
    return tuple([FrameSelect(lambda frames, i=i: frames[i:i - n + 1 + len(frames)], channel) for i in range(n)])



import numpy as np

from phi import math, struct
from phi.geom import Geometry
from phi.physics import State
from phi.physics.field.flag import _PROPAGATOR


def _to_valid_data(data):
    if data is None:
        return None
    if isinstance(data, (tuple, list)):
        return np.array(data)  # numbers or objects
    else:
        return data


@struct.definition()
class Field(State):

    def __init__(self, data, name=None, **kwargs):
        if 'tags' not in kwargs:
            tags = [name, 'field'] if name is not None else ['field']
        State.__init__(self, **struct.kwargs(locals()))

    @property
    def points(self):
        """
        Returns a Field containing all sample points of this field.
        The returned Field is compatible with this one.
        If the components of this field are sampled at different locations, this method raises StaggeredSamplePoints.
        If this field has no sample points, points is None.
        :return: vector Field
        """
        raise NotImplementedError(self)

    @property
    def elements(self):
        """
        Returns the geometry of all cells/particles/elements at which this Field is sampled.
        If the components of this field are sampled at different locations, this method raises StaggeredSamplePoints.
        If this field has no sample points, this method returns None.
        :rtype: Geometry
        :return: all sample elements
        """
        if self.points is None:  # may raise StaggeredSamplePoints
            return None
        raise NotImplementedError(self)

    @struct.variable()
    def data(self, data):
        """
        Data holds the values of this field according to the order specified by points.
        For composite fields, data holds a tuple of component fields.
        :return: n-dimensional tensor
        """
        return _to_valid_data(data)

    @struct.constant()
    def flags(self, flags):
        """
        Flags describe constants_dict of a Field such as divergence-freeness.
        :return: tuple of flags
        """
        if flags is None:
            return ()
        else:
            flags = tuple(set(flags))  # remove duplicates
            for flag in flags:
                if not flag.is_applicable(self.rank, self.component_count):
                    raise ValueError('Flag "%s" is not applicable to field %s' % (flag, self))
            return flags

    def sample_at(self, points):
        """
        Samples this field at the given points.
        :param points: tensor or rank >= 2 containing world-space vectors
        :return: tensor of shape (*location.shape[:-1], field.component_count)
        """
        raise NotImplementedError(self)

    def approximate_mean_value_in(self, geometry):
        """
        Computes the (approximate) mean field value inside the region specified by `geometry`.
        The geometry is assumed to be small compared to the structure of this field and may be approximated as a simpler shape.

        Let V be the volume of `geometry`. This method approximates `1/V * integral_V field.sample_at(x) dx`.
        :param geometry: (batched) Geometry
        :type geometry: Geometry
        :return: float tensor of shape (*geometry.batch_dimensions, field.component_count)
        """
        assert isinstance(geometry, Geometry)
        return self.sample_at(geometry.center)

    def at(self, other_field):
        """
        Resample this field at the same points as other_field.
        The returned Field is compatible with other_field.
        :param other_field: Field
        :return: a new Field which samples all components of this field at the points of other_field
        """
        try:
            values = self.approximate_mean_value_in(other_field.elements)
            result = other_field.copied_with(data=values, flags=propagate_flags_resample(self, other_field.flags, other_field.rank))
            return result
        except StaggeredSamplePoints:  # other_field is staggered
            return broadcast_at(self, other_field)

    @property
    def rank(self):
        """
        Spatial rank of the field (1 for 1D, 2 for 2D, 3 for 3D).
        Note that this does not indicate the shape of the data array.
        If the field is independent of the dimensionality, the rank property is None.
        :return: int
        """
        raise NotImplementedError(self)

    @property
    def component_count(self):
        """
        Number of components of this Field.
        The components can be sampled at the same points or at different points (like with StaggeredGrids).
        :return: int
        :rtype: int
        """
        raise NotImplementedError(self)

    def unstack(self):
        """
        Split the Field by components.
        If the field only has one component, returns a list containing itself.
        :return: tuple of Fields
        :rtype: tuple
        """
        raise NotImplementedError(self)

    @property
    def has_points(self):
        try:
            return self.points is not None
        except StaggeredSamplePoints:
            return True

    def compatible(self, other_field):
        """
        Checks if two Fields have the same sample points and values are stored in the same order.
        For performance reasons, this method does not actually check every single point.
        Even if this method returns False, the sample points may still be the same.
        :type other_field: Field
        :return: True if both Fields have the same sample points.
        :rtype: bool
        """
        raise NotImplementedError(self)

    def with_data(self, data):
        return self.copied_with(data=data, flags=())

    def __mul__(self, other):
        return self.__dataop__(other, True, lambda d1, d2: math.mul(d1, d2))

    __rmul__ = __mul__

    def __truediv__(self, other):
        return self.__dataop__(other, True, lambda d1, d2: math.div(d1, d2))

    def __rtruediv__(self, other):
        return self.__dataop__(other, False, lambda d1, d2: math.div(d2, d1))

    def __sub__(self, other):
        return self.__dataop__(other, False, lambda d1, d2: math.sub(d1, d2))

    def __rsub__(self, other):
        return self.__dataop__(other, False, lambda d1, d2: math.sub(d2, d1))

    def __add__(self, other):
        return self.__dataop__(other, False, lambda d1, d2: math.add(d1, d2))

    __radd__ = __add__

    def __pow__(self, power, modulo=None):
        return self.__dataop__(power, False, lambda f, p: math.pow(f, p))

    def __neg__(self):
        return self * -1

    def __gt__(self, other):
        return self.__dataop__(other, False, lambda x, y: x > y)

    def __ge__(self, other):
        return self.__dataop__(other, False, lambda x, y: x >= y)

    def __lt__(self, other):
        return self.__dataop__(other, False, lambda x, y: x < y)

    def __le__(self, other):
        return self.__dataop__(other, False, lambda x, y: x <= y)

    def __and__(self, other):
        return self.__dataop__(other, False, lambda x, y: x & y)

    def __or__(self, other):
        return self.__dataop__(other, False, lambda x, y: x | y)

    def __dataop__(self, other, linear_if_scalar, data_operator):
        if isinstance(other, Field):
            assert self.compatible(other), 'Fields are not compatible: %s and %s' % (self, other)
            flags = propagate_flags_operation(self.flags + other.flags, False, self.rank, self.component_count)
            self_data = self.data if self.has_points else self.at(other).data
            other_data = other.data if other.has_points else other.at(self).data
            data = data_operator(self_data, other_data)
        else:
            flags = propagate_flags_operation(self.flags, linear_if_scalar, self.rank, self.component_count)
            data = data_operator(self.data, other)
        return self.copied_with(data=data, flags=flags)

    def default_physics(self):
        from .effect import FieldPhysics
        return FieldPhysics(self.name)


class StaggeredSamplePoints(Exception):

    def __init__(self, *args):
        Exception.__init__(self, *args)


class IncompatibleFieldTypes(Exception):
    def __init__(self, *args):
        Exception.__init__(self, *args)


def propagate_flags_resample(data_field, structure_flags, resulting_rank):
    flags = []
    for flag in data_field.flags:
        if flag.is_data_bound and \
                flag.propagates(_PROPAGATOR.RESAMPLE) and \
                flag.is_applicable(resulting_rank, data_field.component_count):
            flags.append(flag)
    for flag in structure_flags:
        if flag.is_structure_bound and \
                flag.propagates(_PROPAGATOR.RESAMPLE) and \
                flag.is_applicable(resulting_rank, data_field.component_count):
            flags.append(flag)
    return tuple(flags)


def propagate_flags_children(flags, child_rank, child_component_count):
    result = []
    for flag in flags:
        if flag.propagates(_PROPAGATOR.CHILDREN) and flag.is_applicable(child_rank, child_component_count):
            result.append(flag)
    return tuple(result)


def propagate_flags_operation(flags, is_linear, result_rank, result_components):
    result = []
    propagator = _PROPAGATOR.LINEAR_OPERATIONS if is_linear else _PROPAGATOR.ALL_OPERATIONS
    for flag in flags:
        if flag.is_data_bound and\
                flag.propagates(propagator) and\
                flag.is_applicable(result_rank, result_components):
            result.append(flag)
    return tuple(result)


def broadcast_at(field1, field2):
    if field1.component_count != field2.component_count and field1.component_count != 1:
        raise IncompatibleFieldTypes('Can only resample to staggered fields with same number of components.\n%s\n%s' % (field1, field2))
    if field1.component_count == 1:
        new_components = [field1.at(f2) for f2 in field2.unstack()]
    else:
        new_components = [f1.at(f2) for f1, f2 in zip(field1.unstack(), field2.unstack())]
    return field2.copied_with(data=tuple(new_components), flags=propagate_flags_resample(field1, field2.flags, field2.rank))



import numbers
import uuid
import warnings
from packaging import version
import six

import numpy as np
import six
import tensorflow as tf
from packaging import version

from phi.backend.backend_helper import split_multi_mode_pad, PadSettings, general_grid_sample_nd, equalize_shapes, circular_pad, replicate_pad
from phi.backend.scipy_backend import SciPyBackend
from phi.tf.tf_cuda_resample import *
from . import tf

from phi.backend.backend import Backend
from phi.backend.tensorop import expand, collapsed_gather_nd


class TFBackend(Backend):

    def __init__(self):
        Backend.__init__(self, "TensorFlow")

    @property
    def precision_dtype(self):
        return {16: np.float16, 32: np.float32, 64: np.float64, None: np.float32}[self.precision]

    def is_tensor(self, x, only_native=False):
        if not only_native and SciPyBackend().is_tensor(x, only_native=False):
            return True
        return isinstance(x, (tf.Tensor, tf.Variable, tf.SparseTensor, tf.Operation))

    def as_tensor(self, x, convert_external=True):
        if self.is_tensor(x, only_native=convert_external):
            tensor = x
        elif isinstance(x, np.ndarray):
            tensor = tf.convert_to_tensor(SciPyBackend(precision=self.precision).as_tensor(x))
        else:
            tensor = tf.convert_to_tensor(x)
        # --- Enforce Precision ---
        if not isinstance(tensor, numbers.Number):
            if isinstance(tensor, np.ndarray):
                tensor = SciPyBackend(precision=self.precision).as_tensor(tensor)
            elif tensor.dtype.is_floating and self.has_fixed_precision:
                tensor = self.to_float(tensor)
        return tensor

    def copy(self, tensor, only_mutable=False):
        if not only_mutable or tf.executing_eagerly():
            return tf.identity(tensor)
        else:
            return tensor

    def equal(self, x, y):
        return tf.equal(x, y)

    def divide_no_nan(self, x, y):
        if version.parse(tf.__version__) >= version.parse('1.11.0'):
            return tf.div_no_nan(x, y)
        else:
            result = x / y
            return tf.where(tf.is_finite(result), result, tf.zeros_like(result))

    def random_uniform(self, shape, low=0, high=1):
        return tf.random.uniform(shape, minval=low, maxval=high, dtype=self.precision_dtype)

    def random_normal(self, shape):
        return tf.random.normal(shape, dtype=self.precision_dtype)

    def rank(self, value):
        return len(value.shape)

    def range(self, start, limit=None, delta=1, dtype=None):
        return tf.range(start, limit, delta, dtype)

    def tile(self, value, multiples):
        if isinstance(multiples, (tuple, list)) and self.ndims(value) < len(multiples):
            value = self.expand_dims(value, axis=0, number=len(multiples) - self.ndims(value))
        return tf.tile(value, multiples)

    def stack(self, values, axis=0):
        return tf.stack(values, axis=axis)

    def concat(self, values, axis):
        return tf.concat(values, axis)

    def pad(self, value, pad_width, mode='constant', constant_values=0):
        passes = split_multi_mode_pad(self.ndims(value), PadSettings(pad_width, mode, constant_values), split_by_constant_value=True)
        for pad_pass in passes:
            value = self._single_mode_single_constant_pad(value, *pad_pass)
        return value

    def _single_mode_single_constant_pad(self, value, pad_width, single_mode, constant_value=0):
        assert single_mode in ('constant', 'symmetric', 'circular', 'reflect', 'replicate'), single_mode
        if single_mode == 'circular':
            return circular_pad(value, pad_width, self)
        if single_mode == 'replicate':
            if np.any(np.array(pad_width) > 1):
                return replicate_pad(value, pad_width, self)
            else:
                single_mode = 'symmetric'
        return tf.pad(value, pad_width, single_mode.upper(), constant_values=constant_value)  # constant, symmetric, reflect

    def reshape(self, value, shape):
        return tf.reshape(value, shape)

    def sum(self, value, axis=None, keepdims=False):
        if axis is not None:
            if not isinstance(axis, int):
                axis = list(axis)
        return tf.reduce_sum(value, axis=axis, keepdims=keepdims)

    def prod(self, value, axis=None):
        if axis is not None:
            if not isinstance(axis, int):
                axis = list(axis)
        if value.dtype == bool:
            return tf.reduce_all(value, axis=axis)
        return tf.reduce_prod(value, axis=axis)

    def where(self, condition, x=None, y=None):
        c = self.cast(condition, self.dtype(x))
        return c * x + (1 - c) * y
        # return tf.where(condition, x, y)  # TF1 has an inconsistent broadcasting rule for where

    def mean(self, value, axis=None, keepdims=False):
        if axis is not None:
            if not isinstance(axis, int):
                axis = list(axis)
        return tf.reduce_mean(value, axis, keepdims=keepdims)

    def py_func(self, func, inputs, Tout, shape_out, stateful=True, name=None, grad=None):
        if grad is None:
            result = tf.py_func(func, inputs, Tout, stateful=stateful, name=name)
        else:
            # Need to generate a unique name to avoid duplicates:
            rnd_name = 'PyFuncGrad' + str(uuid.uuid4())

            tf.RegisterGradient(rnd_name)(grad)  # see _MySquareGrad for grad example
            g = tf.get_default_graph()
            with g.gradient_override_map({"PyFunc": rnd_name}):
                result = tf.py_func(func, inputs, Tout, stateful=stateful, name=name)
        if shape_out is not None:
            result.set_shape(shape_out)
        return result

    def resample(self, inputs, sample_coords, interpolation='linear', boundary='constant', constant_values=0):
        assert interpolation == 'linear'
        if use_cuda(inputs):
            return resample_cuda(inputs, sample_coords, boundary)
        else:
            return general_grid_sample_nd(inputs, sample_coords, boundary, constant_values, self)  # while this is a bit slower than niftynet, it give consisten results at the boundaries

    def zeros_like(self, tensor):
        return tf.zeros_like(tensor)

    def ones_like(self, tensor):
        return tf.ones_like(tensor)

    def dot(self, a, b, axes):
        return tf.tensordot(a, b, axes)

    def matmul(self, A, b):
        if isinstance(A, tf.SparseTensor):
            result = tf.sparse_tensor_dense_matmul(A, tf.transpose(b))
            result = tf.transpose(result)
            result.set_shape(tf.TensorShape([b.shape[0], A.shape[0]]))
            return result
        else:
            return tf.matmul(A, b)

    def einsum(self, equation, *tensors):
        return tf.einsum(equation, *tensors)

    def while_loop(self, cond, body, loop_vars, shape_invariants=None, parallel_iterations=10, back_prop=True,
                   swap_memory=False, name=None, maximum_iterations=None):
        return tf.while_loop(cond, body, loop_vars,
                             shape_invariants=shape_invariants,
                             parallel_iterations=parallel_iterations,
                             back_prop=back_prop,
                             swap_memory=swap_memory,
                             name=name,
                             maximum_iterations=maximum_iterations)

    def abs(self, x):
        return tf.abs(x)

    def sign(self, x):
        return tf.sign(x)

    def round(self, x):
        return tf.round(x)

    def ceil(self, x):
        return tf.ceil(x)

    def floor(self, x):
        return tf.floor(x)

    def max(self, x, axis=None, keepdims=False):
        return tf.reduce_max(x, axis=axis, keepdims=keepdims)

    def min(self, x, axis=None, keepdims=False):
        return tf.reduce_min(x, axis=axis, keepdims=keepdims)

    def with_custom_gradient(self, function, inputs, gradient, input_index=0, output_index=None, name_base="custom_gradient_func"):
        # Setup custom gradient
        gradient_name = name_base + "_" + str(uuid.uuid4())
        tf.RegisterGradient(gradient_name)(gradient)

        g = tf.get_default_graph()
        with g.gradient_override_map({"Identity": gradient_name}):
            fake_function = tf.identity(inputs[input_index])

        outputs = function(*inputs)
        output = outputs if output_index is None else outputs[output_index]
        output_with_gradient = fake_function + tf.stop_gradient(output - fake_function)
        if output_index is None:
            return output_with_gradient
        else:
            outputs = list(outputs)
            outputs[output_index] = output_with_gradient
            return outputs

    def maximum(self, a, b):
        return tf.maximum(a, b)

    def minimum(self, a, b):
        return tf.minimum(a, b)

    def clip(self, x, minimum, maximum):
        return tf.clip_by_value(x, minimum, maximum)

    def sqrt(self, x):
        return tf.sqrt(x)

    def exp(self, x):
        return tf.exp(x)

    def conv(self, tensor, kernel, padding="SAME"):
        rank = tensor_spatial_rank(tensor)
        padding = padding.upper()
        if rank == 1:
            result = tf.nn.conv1d(tensor, kernel, 1, padding)
        elif rank == 2:
            result = tf.nn.conv2d(tensor, kernel, [1, 1, 1, 1], padding)
        elif rank == 3:
            result = tf.nn.conv3d(tensor, kernel, [1, 1, 1, 1, 1], padding)
        else:
            raise ValueError("Tensor must be of rank 1, 2 or 3 but is %d" % rank)
        return result

    def expand_dims(self, a, axis=0, number=1):
        if number == 0:
            return a
        for _i in range(number):
            a = tf.expand_dims(a, axis)
        return a

    def shape(self, tensor):
        return tf.shape(tensor)

    def to_float(self, x, float64=False):
        if float64:
            warnings.warn('float64 argument is deprecated, set Backend.precision = 64 to use 64 bit operations.', DeprecationWarning)
            return tf.cast(x, tf.float64)
        else:
            return tf.cast(x, self.precision_dtype)

    def staticshape(self, tensor):
        if self.is_tensor(tensor, only_native=True):
            return tuple(tensor.shape.as_list())
        else:
            return np.shape(tensor)

    def to_int(self, x, int64=False):
        return tf.cast(x, tf.int64) if int64 else tf.cast(x, tf.int32)

    def to_complex(self, x):
        if self.dtype(x) in (np.complex64, np.complex128):
            return x
        if self.dtype(x) == np.float64:
            return tf.to_complex128(x)
        else:
            return tf.to_complex64(x)

    def gather(self, values, indices):
        if isinstance(indices, slice):
            return values[indices]
        return tf.gather(values, indices)

    def gather_nd(self, values, indices, batch_dims=0):
        if batch_dims == 0:
            return tf.gather_nd(values, indices)
        elif version.parse(tf.__version__) >= version.parse('1.14.0'):
            return tf.gather_nd(values, indices, batch_dims=batch_dims)
        else:
            if batch_dims > 1:
                raise NotImplementedError('batch_dims > 1 only supported on TensorFlow >= 1.14')
            batch_size = self.shape(values)[0]
            batch_ids = tf.reshape(tf.range(batch_size), [batch_size] + [1] * (self.ndims(indices) - 1))
            batch_ids = tf.tile(batch_ids, [1] + self.shape(indices)[1:-1] + [1])
            indices = tf.concat([batch_ids, indices], -1)
            return tf.gather_nd(values, indices)

    def unstack(self, tensor, axis=0, keepdims=False):
        unstacked = tf.unstack(tensor, axis=axis)
        if keepdims:
            unstacked = [self.expand_dims(c, axis=axis) for c in unstacked]
        return unstacked

    def std(self, x, axis=None, keepdims=False):
        _mean, var = tf.nn.moments(x, axis, keepdims=keepdims)
        return tf.sqrt(var)

    def boolean_mask(self, x, mask):
        return tf.boolean_mask(x, mask)

    def isfinite(self, x):
        return tf.is_finite(x)

    def any(self, boolean_tensor, axis=None, keepdims=False):
        return tf.reduce_any(boolean_tensor, axis=axis, keepdims=keepdims)

    def all(self, boolean_tensor, axis=None, keepdims=False):
        return tf.reduce_all(boolean_tensor, axis=axis, keepdims=keepdims)

    def scatter(self, points, indices, values, shape, duplicates_handling='undefined'):
        # Change indexing so batch number is included as first element of the index, for example: [0,31,24] indexes the first batch (batch 0) and 2D coordinates (31,24).
        buffer = tf.zeros(shape, dtype=values.dtype)

        repetitions = []
        for dim in range(len(indices.shape) - 1):
            if values.shape[dim] == 1:
                repetitions.append(indices.shape[dim])
            else:
                assert indices.shape[dim] == values.shape[dim]
                repetitions.append(1)
        repetitions.append(1)
        values = self.tile(values, repetitions)

        if duplicates_handling == 'add':
            # Only for Tensorflow with custom gradient
            @tf.custom_gradient
            def scatter_density(points, indices, values):
                result = tf.tensor_scatter_add(buffer, indices, values)

                def grad(dr):
                    return self.resample(gradient(dr, difference='central'), points), None, None

                return result, grad

            return scatter_density(points, indices, values)
        elif duplicates_handling == 'mean':
            # Won't entirely work with out of bounds particles (still counted in mean)
            count = tf.tensor_scatter_add(buffer, indices, tf.ones_like(values))
            total = tf.tensor_scatter_add(buffer, indices, values)
            return total / tf.maximum(1.0, count)
        else:  # last, any, undefined
            # indices = self.to_int(indices, int64=True)
            # st = tf.SparseTensor(indices, values, shape)  # ToDo this only supports 2D shapes
            # st = tf.sparse.reorder(st)   # only needed if not ordered
            # return tf.sparse.to_dense(st)
            count = tf.tensor_scatter_add(buffer, indices, tf.ones_like(values))
            total = tf.tensor_scatter_add(buffer, indices, values)
            return total / tf.maximum(1.0, count)

    def fft(self, x):
        rank = len(x.shape) - 2
        assert rank >= 1
        x = self.to_complex(x)
        if rank == 1:
            return tf.stack([tf.fft(c) for c in tf.unstack(x, axis=-1)], axis=-1)
        elif rank == 2:
            return tf.stack([tf.fft2d(c) for c in tf.unstack(x, axis=-1)], axis=-1)
        elif rank == 3:
            return tf.stack([tf.fft3d(c) for c in tf.unstack(x, axis=-1)], axis=-1)
        else:
            raise NotImplementedError('n-dimensional FFT not implemented.')

    def ifft(self, k):
        rank = len(k.shape) - 2
        assert rank >= 1
        if rank == 1:
            return tf.stack([tf.ifft(c) for c in tf.unstack(k, axis=-1)], axis=-1)
        elif rank == 2:
            return tf.stack([tf.ifft2d(c) for c in tf.unstack(k, axis=-1)], axis=-1)
        elif rank == 3:
            return tf.stack([tf.ifft3d(c) for c in tf.unstack(k, axis=-1)], axis=-1)
        else:
            raise NotImplementedError('n-dimensional inverse FFT not implemented.')

    def imag(self, complex):
        return tf.imag(complex)

    def real(self, complex):
        return tf.real(complex)

    def cast(self, x, dtype):
        return tf.cast(x, dtype)

    def sin(self, x):
        return tf.sin(x)

    def cos(self, x):
        return tf.cos(x)

    def dtype(self, array):
        if self.is_tensor(array, only_native=True):
            return array.dtype.as_numpy_dtype
        else:
            return SciPyBackend().dtype(array)

    def sparse_tensor(self, indices, values, shape):
        return tf.SparseTensor(indices=indices, values=values, dense_shape=shape)


# from niftynet.layer.resampler.py
# https://cmiclab.cs.ucl.ac.uk/CMIC/NiftyNet/blob/69c98e5a95cc6788ad9fb8c5e27dc24d1acec634/niftynet/layer/resampler.py


COORDINATES_TYPE = tf.int32
EPS = 1e-6


def tensor_spatial_rank(tensor):
    return len(tensor.shape) - 2


def unit_direction(dim, spatial_rank):  # ordered like z,y,x
    direction = [1 if i == dim else 0 for i in range(spatial_rank)]
    for _i in range(spatial_rank):
        direction = tf.expand_dims(direction, axis=0)
    return direction


def _resample_no_pack(grid, coords, boundary_func):
    resolution = np.array([int(d) for d in grid.shape[1:-1]])
    sp_rank = tensor_spatial_rank(grid)

    floor = boundary_func(tf.floor(coords), resolution)
    up_weights = coords - floor
    lo_weights = TFBackend().unstack(1 - up_weights, axis=-1, keepdims=True)
    up_weights = TFBackend().unstack(up_weights, axis=-1, keepdims=True)
    base_coords = tf.cast(floor, tf.int32)

    def interpolate_nd(coords, axis):
        direction = np.array([1 if ax == axis else 0 for ax in range(sp_rank)])
        print(direction.shape)
        with tf.variable_scope('coord_plus_one'):
            up_coords = coords + direction  # This is extremely slow for some reason - ToDo tile direction array to have same dimensions before calling interpolate_nd?
        if axis == sp_rank - 1:
            # up_coords = boundary_func(up_coords, resolution)
            lo_values = tf.gather_nd(grid, coords, batch_dims=1)
            up_values = tf.gather_nd(grid, up_coords, batch_dims=1)
        else:
            lo_values = interpolate_nd(coords, axis + 1)
            up_values = interpolate_nd(up_coords, axis + 1)
        with tf.variable_scope('weighted_sum_axis_%d' % axis):
            return lo_values * lo_weights[axis] + up_values * up_weights[axis]

    with tf.variable_scope('interpolate_nd'):
        result = interpolate_nd(base_coords, 0)
    return result


def _resample_linear_niftynet(inputs, sample_coords, boundary, boundary_func, float_type):
    inputs = tf.convert_to_tensor(inputs)
    sample_coords = tf.convert_to_tensor(sample_coords)

    in_spatial_size = [int(d) for d in inputs.shape[1:-1]]
    in_spatial_rank = tensor_spatial_rank(inputs)
    batch_size = tf.shape(inputs)[0]

    out_spatial_rank = tensor_spatial_rank(sample_coords)
    out_spatial_size = sample_coords.get_shape().as_list()[1:-1]

    if sample_coords.shape[0] != inputs.shape[0]:
        sample_coords = tf.tile(sample_coords, [batch_size] + [1] * (len(sample_coords.shape) - 1))

    xy = tf.unstack(sample_coords, axis=-1)
    base_coords = [tf.floor(coords) for coords in xy]
    floor_coords = [tf.cast(boundary_func(x, in_spatial_size[idx]), COORDINATES_TYPE) for (idx, x) in enumerate(base_coords)]
    ceil_coords = [tf.cast(boundary_func(x + 1.0, in_spatial_size[idx]), COORDINATES_TYPE) for (idx, x) in enumerate(base_coords)]

    if boundary.upper() == 'ZERO':
        weight_0 = [tf.expand_dims(x - tf.cast(i, float_type), -1) for (x, i) in zip(xy, floor_coords)]
        weight_1 = [tf.expand_dims(tf.cast(i, float_type) - x, -1) for (x, i) in zip(xy, ceil_coords)]
    else:
        weight_0 = [tf.expand_dims(x - i, -1) for (x, i) in zip(xy, base_coords)]
        weight_1 = [1.0 - w for w in weight_0]

    batch_ids = tf.reshape(tf.range(batch_size), [batch_size] + [1] * out_spatial_rank)
    batch_ids = tf.tile(batch_ids, [1] + out_spatial_size)
    sc = (floor_coords, ceil_coords)
    binary_neighbour_ids = [[int(c) for c in format(i, '0%ib' % in_spatial_rank)] for i in range(2 ** in_spatial_rank)]

    def get_knot(bc):
        coord = [sc[c][i] for i, c in enumerate(bc)]
        if version.parse(tf.__version__) >= version.parse('1.14.0'):
            coord = tf.stack(coord, -1)
            return tf.gather_nd(inputs, coord, batch_dims=1)  # NaN can cause negative integers here
        else:
            coord = tf.stack([batch_ids] + coord, -1)
            return tf.gather_nd(inputs, coord)  # NaN can cause negative integers here

    samples = [get_knot(bc) for bc in binary_neighbour_ids]

    def _pyramid_combination(samples, w_0, w_1):
        if len(w_0) == 1:
            return samples[0] * w_1[0] + samples[1] * w_0[0]
        f_0 = _pyramid_combination(samples[::2], w_0[:-1], w_1[:-1])
        f_1 = _pyramid_combination(samples[1::2], w_0[:-1], w_1[:-1])
        return f_0 * w_1[-1] + f_1 * w_0[-1]

    return _pyramid_combination(samples, weight_0, weight_1)


def _boundary_snap(sample_coords, spatial_shape):
    max_indices = [l - 1 for l in spatial_shape]
    for _i in range(len(spatial_shape)):
        max_indices = tf.expand_dims(max_indices, 0)
    sample_coords = tf.minimum(sample_coords, max_indices)
    sample_coords = tf.maximum(sample_coords, 0)
    return sample_coords


def _boundary_replicate(sample_coords, input_size):
    return tf.maximum(tf.minimum(sample_coords, input_size - 1), 0)


def _boundary_circular(sample_coords, input_size):
    return tf.mod(tf.mod(sample_coords, input_size) + input_size, input_size)


def _boundary_symmetric(sample_coords, input_size):
    sample_coords = _boundary_circular(sample_coords, 2 * input_size)
    return ((2 * input_size - 1) - tf.abs((2 * input_size - 1) - 2 * sample_coords)) // 2


def _boundary_reflect(sample_coords, input_size):
    sample_coords = _boundary_circular(sample_coords, 2 * input_size - 2)
    return (input_size - 1) - tf.abs((input_size - 1) - sample_coords)


SUPPORTED_BOUNDARY = {
    'zero': _boundary_replicate,
    'replicate': _boundary_replicate,
    'circular': _boundary_circular,
    'symmetric': _boundary_symmetric,
    'reflect': _boundary_reflect,
}



import six
from collections import namedtuple

import numpy as np

from .tensorop import expand, collapsed_gather_nd, CollapsedTensor as CT, collapse


PadSettings = namedtuple('PadSettings', ['pad_width', 'mode', 'constant_values'])


def split_multi_mode_pad(tensor_rank, pad_settings, split_by_constant_value=True):
    dims = range(tensor_rank)
    pad_width, mode, constant_values = pad_settings
    if isinstance(mode, six.string_types):
        if not split_by_constant_value or not isinstance(constant_values, (tuple, list)):
            if isinstance(constant_values, (tuple, list)):
                constant_values = expand(constant_values, shape=(len(dims), 2))
            pad_width = expand(pad_width, [tensor_rank, 2])
            return [PadSettings(pad_width, mode, constant_values)]
    mode = expand(mode, shape=(len(dims), 2))
    constant_values = expand(constant_values, shape=(len(dims), 2))
    passes = [('circular', 0), ('wrap', 0), ('replicate', 0), ('symmetric', 0), ('reflect', 0)]
    if split_by_constant_value:
        constant_value_set = set()
        for dim in dims:
            for upper in (False, True):
                constant_value_set.add(constant_values[dim][upper])
        for const in constant_value_set:
            passes.append(('constant', const))
    else:
        passes.append(('constant', constant_values))
    result = []  # list of PadSettings
    for single_mode, constant_value in passes:  # order matters! circular/wrap first
        widths = [[collapsed_gather_nd(pad_width, [dim, upper]) if mode[dim][upper] == single_mode and constant_values[dim][upper] == collapsed_gather_nd(constant_value, [dim, upper]) else 0 for upper in (False, True)] for dim in dims]
        if np.sum(np.array(widths)) > 0:
            result.append(PadSettings(widths, single_mode, constant_value))
    if np.sum(np.array(pad_width)) > 0 and len(result) == 0:
        split_multi_mode_pad(tensor_rank, pad_settings, split_by_constant_value=split_by_constant_value)

    return result


NeighbourReduce = namedtuple('NeighbourReduce', ['requires_weights', 'f'])


def general_grid_sample_nd(grid, coords, boundary, constant_values, math, reduce='linear'):
    """
    Backend-independent grid sampling with linear interpolation.
    Supports boundary conditions per face: 'constant' , 'replicate', 'circular', 'symmetric', 'reflect'.

    Interpolation at the boundaries works according to the following principle:
    The boundary mode determines the value at virtual grid points outside the grid bounds.
    This is exact, even for far-away points.
    Then, linear interpolation is used to determine the point between grid points.
    Consequently, for constant boundaries, the value linearly approaches the constant value over the distance of one cell at the boundary.

    :param grid: tensor of shape (batch_dim, spatial dims..., channels)
    :param coords: tensor of shape (batch_dim, ..., spatial_rank)
    :param boundary: 'zero'/'constant', 'replicate', 'circular', 'symmetric', 'reflect'
    :param constant_values: extrapolation values (same options as in pad)
    :param math: backend
    :return: tensor of sampled values from the grid
    """
    if not isinstance(reduce, NeighbourReduce):
        reduce = {
            'linear': NeighbourReduce(True, lambda v1, v2, w1, w2: v1 * w1 + v2 * w2),
            'min': NeighbourReduce(False, lambda v1, v2: math.minimum(v1, v2)),
            'max': NeighbourReduce(False, lambda v1, v2: math.maximum(v1, v2)),
            'minmax': NeighbourReduce(False, lambda v1, v2: (math.minimum(v1[0], v2[0]), math.maximum(v1[1], v2[1])) if isinstance(v1, tuple) else (math.minimum(v1, v2), math.maximum(v1, v2))),
        }[reduce]
    grid, coords, boundary = pad_constant_boundaries(grid, coords, boundary, constant_values, math)

    resolution = np.array([int(d) for d in grid.shape[1:-1]])
    sp_rank = math.ndims(grid) - 2
    # --- Compute weights ---
    floor = math.floor(coords)
    lo_coords = math.to_int(floor)
    hi_coords = apply_boundary(boundary, lo_coords + 1, resolution, math)
    lo_coords = apply_boundary(boundary, lo_coords, resolution, math)
    if reduce.requires_weights:
        hi_weights = coords - floor
        lo_weights = math.unstack(1 - hi_weights, axis=-1, keepdims=True)
        hi_weights = math.unstack(hi_weights, axis=-1, keepdims=True)

    def interpolate_nd(is_hi_by_axis, axis):
        is_hi_by_axis_2 = is_hi_by_axis | np.array([ax == axis for ax in range(sp_rank)])
        coords1 = math.where(is_hi_by_axis, hi_coords, lo_coords)
        coords2 = math.where(is_hi_by_axis_2, hi_coords, lo_coords)
        if axis == sp_rank - 1:
            lo_values = math.gather_nd(grid, coords1, batch_dims=1)
            hi_values = math.gather_nd(grid, coords2, batch_dims=1)
        else:
            lo_values = interpolate_nd(is_hi_by_axis, axis + 1)
            hi_values = interpolate_nd(is_hi_by_axis_2, axis + 1)
        if reduce.requires_weights:
            return reduce.f(lo_values, hi_values, lo_weights[axis], hi_weights[axis])
        else:
            return reduce.f(lo_values, hi_values)
    result = interpolate_nd(np.array([False] * sp_rank), 0)
    return result


def pad_constant_boundaries(grid, coords, boundary, constant_values, math):
    boundary = CT(boundary)
    spatial_rank = math.staticshape(coords)[-1]
    pad_widths = [[1 if boundary[dim, upper] in ('zero', 'constant') else 0 for upper in (False, True)] for dim in range(-spatial_rank - 1, -1)]
    boundary = [['replicate' if boundary[dim, upper] in ('zero', 'constant') else boundary[dim, upper] for upper in (False, True)] for dim in range(-spatial_rank - 1, -1)]
    lower_pads = [lu[0] for lu in pad_widths]
    grid = math.pad(grid, [[0, 0]] + pad_widths + [[0, 0]], mode='constant', constant_values=constant_values)
    if sum(lower_pads) > 0:
        coords = math.add(coords, math.cast(lower_pads, math.dtype(coords)))
    boundary = collapse(boundary)
    return grid, coords, boundary


def apply_boundary(boundary, coords, input_size, math):
    if isinstance(boundary, six.string_types):
        return _apply_single_boundary(boundary, coords, input_size, math)
    coords = math.unstack(coords, axis=-1)
    assert len(boundary) == len(input_size) == len(coords)
    boundary = CT(boundary)
    result = []
    for dim, dim_coords in enumerate(coords):
        if boundary[dim, 0] == boundary[dim, 1]:
            result.append(_apply_single_boundary(boundary[dim, 0], dim_coords, input_size[dim], math))
        else:  # separate boundary for lower and upper face
            lower = _apply_single_boundary(boundary[dim, 0], dim_coords, input_size[dim], math)
            upper = _apply_single_boundary(boundary[dim, 1], dim_coords, input_size[dim], math)
            result.append(math.where(dim_coords <= 0, lower, upper))
    return math.stack(result, axis=-1)


def _apply_single_boundary(boundary, coords, input_size, math):
    if boundary == 'zero' or boundary == 'constant':
        raise ValueError("boundary 'zero' cannot be applied to coordinates")
    elif boundary == 'replicate':
        return math.clip(coords, 0, input_size - 1)
    elif boundary == 'circular':
        return math.mod(coords, input_size)
    elif boundary == 'symmetric':
        coords = math.mod(coords, 2 * input_size)
        return ((2 * input_size - 1) - math.abs((2 * input_size - 1) - 2 * coords)) // 2
    elif boundary == 'reflect':
        coords = math.mod(coords, 2 * input_size - 2)
        return (input_size - 1) - math.abs((input_size - 1) - coords)
    else:
        raise ValueError('Invalid boundary: %s' % (boundary,))


def equalize_ranks(tensors, math):
    rank = max([math.ndims(tensor) for tensor in tensors])
    return [math.expand_dims(tensor, 0, number=rank - math.ndims(tensor)) for tensor in tensors]


def equalize_shapes(tensors, math, ignore_outer_dims=0):
    tensors = equalize_ranks(tensors, math)
    shape = _combined_shape([math.staticshape(t) for t in tensors], ignore_outer_dims)
    result = []
    for tensor in tensors:
        tensor_shape = math.staticshape(tensor)
        multiples = [1] * ignore_outer_dims
        for dim in range(ignore_outer_dims, len(tensor_shape)):
            multiples.append(shape[dim] // tensor_shape[dim])
        result.append(math.tile(tensor, multiples))
    return result


def _combined_shape(shapes, ignore_outer_dims):
    rank = len(shapes[0])  # assume all shapes have same rank
    resulting_shape = [None] * ignore_outer_dims
    for dim in range(ignore_outer_dims, rank):
        dim_value = None
        for shape in shapes:
            dim_value = combined_dim(dim_value, shape[dim])
        resulting_shape.append(dim_value)
    return tuple(resulting_shape)


def combined_dim(dim1, dim2):
    if dim1 is None and dim2 is None:
        return None
    if dim1 is None or dim1 == 1:
        return dim2
    if dim2 is None or dim2 == 1:
        return dim1
    assert dim1 == dim2, "Cannot bring shapes together because dimensions are incompatible: %d and %d" % (dim1, dim2)
    return dim1


def circular_pad(value, pad_width, math):
    dims = range(math.ndims(value))
    for dim in dims:
        pad_lower, pad_upper = pad_width[dim]
        if pad_lower == 0 and pad_upper == 0:
            continue  # Nothing to pad
        lower = value[tuple([slice(value.shape[dim] - pad_lower, None) if d == dim else slice(None) for d in dims])]
        upper = value[tuple([slice(None, pad_upper) if d == dim else slice(None) for d in dims])]
        value = math.concat([lower, value, upper], axis=dim)
    return value


def replicate_pad(value, pad_width, math):
    dims = range(math.ndims(value))
    for dim in dims:
        pad_lower, pad_upper = pad_width[dim]
        if pad_lower == 0 and pad_upper == 0:
            continue  # Nothing to pad
        bottom_row = value[(slice(None),) + tuple([slice(1) if d == dim else slice(None) for d in dims]) + (slice(None),)]
        top_row = value[(slice(None),) + tuple([slice(-1, None) if d == dim else slice(None) for d in dims]) + (slice(None),)]
        value = math.concat([bottom_row] * pad_lower + [value] + [top_row] * pad_upper)
    return value


def symmetric_pad(value, pad_width, math):
    raise NotImplementedError()  # only used by PyTorch which does not support ::-1 axis flips
    dims = range(math.ndims(value))
    for dim in dims:
        pad_lower, pad_upper = pad_width[dim]
        if pad_lower == 0 and pad_upper == 0:
            continue  # Nothing to pad
        top_rows = value[tuple([slice(value.shape[dim] - pad_upper, None) if d == dim else slice(None) for d in dims])]
        bottom_rows = value[tuple([slice(None, pad_lower) if d == dim else slice(None) for d in dims])]
        top_rows = math.flip_axis(top_rows, dim)
        bottom_rows = math.flip_axis(bottom_rows, dim)
        value = math.concat([bottom_rows, value, top_rows], axis=dim)
    return value



import contextlib
import logging
import os
import threading

import numpy as np
from . import tf
from phi import struct
from .profiling import Timeliner
from .util import isplaceholder, istensor


class Session(object):

    def __init__(self, scene, session=None):
        self._scene = scene
        self._session = session if session is not None else tf.Session()
        assert self._session.graph == tf.get_default_graph(), 'Session %s does not reference the current TensorFlow graph.'
        self.graph = tf.get_default_graph()
        self.summary_writers = {}
        self.summary_directory = os.path.abspath(scene.subpath('summary')) if scene is not None else None
        self.profiling_directory = scene.subpath("profile") if scene is not None else None
        self.trace_count = 0
        self.saver = None

    def initialize_variables(self):
        with self.graph.as_default():
            self._session.run(tf.global_variables_initializer())
            self.saver = tf.train.Saver(max_to_keep=100, allow_empty=True)

    def run(self, fetches, feed_dict=None, summary_key=None, time=None, merged_summary=None, item_condition=struct.ALL_ITEMS):
        if isinstance(fetches, np.ndarray):
            return fetches
        if fetches is None:
            return None

        tensor_feed_dict = None
        if feed_dict is not None:
            tensor_feed_dict = {}
            for (key, value) in feed_dict.items():
                pairs = struct.zip([key, value], item_condition=item_condition, zip_parents_if_incompatible=True)

                def add_to_dict(key_tensor, value_tensor):
                    if isplaceholder(key_tensor):
                        tensor_feed_dict[key_tensor] = value_tensor
                    return None
                struct.map(add_to_dict, pairs, item_condition=item_condition, content_type=struct.INVALID)

        tensor_fetches = struct.flatten(fetches, item_condition=item_condition)
        if isinstance(fetches, (tuple, list)):
            def is_fetch(x): return istensor(x) or _identity_in(x, fetches)
        else:
            def is_fetch(x): return istensor(x) or x is fetches
        tensor_fetches = tuple(filter(is_fetch, tensor_fetches))

        # Handle tracing
        trace = _trace_stack.get_default(raise_error=False)
        if trace:
            options = trace.timeliner.options
            run_metadata = trace.timeliner.run_metadata
        else:
            options = None
            run_metadata = None

        # Summary
        if summary_key is not None and merged_summary is not None:
            tensor_fetches = (merged_summary,) + tensor_fetches

        result_fetches = self._session.run(tensor_fetches, tensor_feed_dict, options, run_metadata)
        result_dict = {fetch: result for fetch, result in zip(tensor_fetches, result_fetches)}

        if summary_key:
            summary_buffer = result_fetches[0]
            result_fetches = result_fetches[1:]
            if summary_key in self.summary_writers:
                summary_writer = self.summary_writers[summary_key]
            else:
                summary_writer = tf.summary.FileWriter(os.path.join(self.summary_directory, str(summary_key)), self.graph)
                self.summary_writers[summary_key] = summary_writer
            summary_writer.add_summary(summary_buffer, time)
            summary_writer.flush()

        if trace:
            trace.timeliner.add_run()

        def replace_tensor_with_value(fetch):
            try:
                if fetch in result_dict:
                    return result_dict[fetch]
                else:
                    return fetch
            except TypeError:  # not hashable
                return fetch
        result = struct.map(replace_tensor_with_value, fetches, item_condition=item_condition)
        return result

    def profiler(self):
        os.path.isdir(self.profiling_directory) or os.makedirs(self.profiling_directory)
        self.trace_count += 1
        return Trace(self.trace_count, self.profiling_directory)

    def save(self, dir):
        assert self.saver is not None, "save() called before initialize_variables()"
        os.path.isdir(dir) or os.makedirs(dir)
        self.saver.save(self._session, os.path.join(dir, "model.ckpt"))

    def restore(self, dir, scope=None):
        path = os.path.join(dir, "model.ckpt")
        vars = self.graph.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=scope)
        if len(vars) == 0:
            raise ValueError('The current graph does not contain any variables in scope "%s.\nAll: %s"' % (scope, self.graph.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)))
        saver = tf.train.Saver(var_list=vars)
        saver.restore(self._session, path)

    def restore_new_scope(self, dir, saved_scope, tf_scope):
        var_remap = dict()
        vars = [v for v in self.graph.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=tf_scope) if "Adam" not in v.name]
        for var in vars:
            var_remap[saved_scope + var.name[len(tf_scope):-2]] = var
        path = os.path.join(dir, "model.ckpt")
        saver = tf.train.Saver(var_list=var_remap)
        try:
            saver.restore(self._session, path)
        except tf.errors.NotFoundError as e:
            from tensorflow.contrib.framework.python.framework import checkpoint_utils
            logging.info(checkpoint_utils.list_variables(dir))
            raise e

    def as_default(self):
        return self._session.as_default()


class Trace(object):

    def __init__(self, index, directory):
        self.index = index
        self.directory = directory
        self.timeliner = None
        self.timeline_file = None
        self._default_simulation_context_manager = None

    def __enter__(self):
        self.timeline_file = os.path.join(self.directory, 'trace %d.json' % self.index)
        self.timeliner = Timeliner()

        if self._default_simulation_context_manager is None:
            self._default_simulation_context_manager = _trace_stack.get_controller(self)
        return self._default_simulation_context_manager.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.timeliner.save(self.timeline_file)

        self._default_simulation_context_manager.__exit__(exc_type, exc_val, exc_tb)
        self._default_simulation_context_manager = None


class _TraceStack(threading.local):

    def __init__(self):
        self.stack = []

    def get_default(self, raise_error=True):
        if raise_error:
            assert len(self.stack) > 0, "Default simulation required. Use 'with simulation:' or 'with simulation.as_default():"
        return self.stack[-1] if len(self.stack) >= 1 else None

    def reset(self):
        self.stack = []

    def is_cleared(self):
        return not self.stack

    @contextlib.contextmanager
    def get_controller(self, default):
        """Returns a context manager for manipulating a default stack."""
        try:
            self.stack.append(default)
            yield default
        finally:
            # stack may be empty if reset() was called
            if self.stack:
                self.stack.remove(default)


_trace_stack = _TraceStack()


def _identity_in(obj, list):
    for item in list:
        if item is obj:
            return True
    return False



import warnings
from numbers import Number
import numpy as np

from phi import struct
from phi.backend.dynamic_backend import DYNAMIC_BACKEND as math
from phi.backend.dynamic_backend import NoBackendFound
from phi.struct.functions import mappable

from .nd import fftfreq


@mappable(item_condition=struct.ALL_ITEMS, content_type=type)
def types(x):
    warnings.warn("math.types is deprecated. Use struct.dtype isntead.", DeprecationWarning)
    try:
        return math.dtype(x)
    except NoBackendFound:
        return type(x)


def is_static_shape(obj):
    if not isinstance(obj, (tuple, list, np.ndarray)):
        return False
    for element in obj:
        if not isinstance(element, Number) and element is not None:
            return False
    return True


def _none_to_one(shape):
    result = list(map(lambda val: 1 if val is None else val, shape))
    return result


@mappable(leaf_condition=is_static_shape)
def zeros(shape, dtype=None):
    if dtype is not None:
        return np.zeros(_none_to_one(shape), dtype=dtype)
    else:
        return math.to_float(np.zeros(_none_to_one(shape), np.int8))


@mappable(leaf_condition=is_static_shape)
def ones(shape, dtype=None):
    if dtype is not None:
        return np.ones(_none_to_one(shape), dtype)
    else:
        return math.to_float(np.ones(_none_to_one(shape), np.int8))


@mappable(leaf_condition=is_static_shape)
def randn(shape, dtype=None):
    array = np.random.randn(*_none_to_one(shape))
    if dtype is not None:
        return array.astype(dtype)
    else:
        return math.to_float(array)


def randfreq(shape, dtype=None, power=8):
    warnings.warn('randfreq() is deprecated. Use Noise() instead.')

    def genarray(shape):
        fft = randn(shape, dtype) + 1j * randn(shape, dtype)
        k = fftfreq(shape[1:-1], mode='absolute')
        shape_fac = math.sqrt(math.mean(shape[1:-1]))  # 16: 4, 64: 8, 256: 24,
        fft *= (1 / (k + 1)) ** power * power * shape_fac
        array = math.real(math.ifft(fft))
        if dtype is not None:
            return array.astype(dtype)
        else:
            return math.to_float(array)
    return struct.map(genarray, shape, leaf_condition=is_static_shape)


def interpolate(alpha, x, y, clip_alpha=True):
    """
    Linear interpolation between x and y according to the value of `alpha`

    :param alpha: blending between `x` (`alpha=0`) and `y` (`alpha=1`)
    :param x: value for `alpha=0`, must be compatible with y and alpha
    :param y: value for `alpha=1`, must be compatible with x and alpha
    :param clip_alpha:
    :return:
    """
    if clip_alpha:
        alpha = math.clip(alpha, 0, 1)
    return x * (1 - alpha) + y * alpha



""" High-level world utility and convenience functions. """
from phi.geom import union
from .world import StateProxy, World
from .field.mask import mask


def obstacle_mask(world_or_proxy, **mask_kwargs):
    """
Builds a binary Field, masking all obstacles in the world.
    :param world_or_proxy: World or StateProxy object
    :return: Field
    """
    world = world_or_proxy.world if isinstance(world_or_proxy, StateProxy) else world_or_proxy
    assert isinstance(world, World)
    geometries = [obstacle.geometry for obstacle in world.state.all_with_tag('obstacle')]
    return mask(union(geometries), **mask_kwargs)



import numpy as np

from phi import struct, math
from phi.geom import Sphere
from phi.physics.domain import Domain

from .field import Field
from .grid import CenteredGrid
from .staggered_grid import StaggeredGrid, unstack_staggered_tensor
from .flag import SAMPLE_POINTS
from .util import extrapolate


@struct.definition()
class SampledField(Field):

    def __init__(self, sample_points, data=1, mode='mean', point_count=None, **kwargs):
        Field.__init__(self, **struct.kwargs(locals(), ignore=['point_count']))
        self._point_count = point_count

    def sample_at(self, points):
        raise NotImplementedError()

    def at(self, other_field):
        if isinstance(other_field, SampledField) and other_field.sample_points is self.sample_points:
            return self
        elif isinstance(other_field, Domain):
            return self._grid_sample(other_field.box, other_field.resolution, 1)
        elif isinstance(other_field, CenteredGrid):
            return self._grid_sample(other_field.box, other_field.resolution, other_field._batch_size)
        elif isinstance(other_field, StaggeredGrid):
            return self._stagger_sample(other_field.box, other_field.resolution)

        else:
            return self

    def _grid_sample(self, box, resolution, batch_size):
        """
    Samples this field on a regular grid.
        :param box: physical dimensions of the grid
        :param resolution: grid resolution
        :return: CenteredGrid
        """
        sample_indices_nd = math.to_int(math.round(box.global_to_local(self.sample_points) * resolution))
        sample_indices_nd = math.minimum(math.maximum(0, sample_indices_nd), resolution - 1)  # Snap outside points to edges, otherwise scatter raises an error
        # Correct format for math.scatter
        valid_indices = _batch_indices(sample_indices_nd)
        if batch_size is None:
            batch_size = self._batch_size
        if batch_size is None:
            batch_size = 1
        shape = (batch_size,) + tuple(resolution) + (self.data.shape[-1],)
        scattered = math.scatter(self.sample_points, valid_indices, self.data, shape, duplicates_handling=self.mode)
        return CenteredGrid(data=scattered, box=box, extrapolation='constant', name=self.name + '_centered')

    def _stagger_sample(self, box, resolution):
        """
    Samples this field on a staggered grid.
    In addition to sampling, extrapolates the field using an occupancy mask generated from the points.
        :param box: physical dimensions of the grid
        :param resolution: grid resolution
        :return: StaggeredGrid
        """
        resolution = np.array(resolution)
        valid_indices = math.to_int(math.floor(self.sample_points))
        valid_indices = math.minimum(math.maximum(0, valid_indices), resolution - 1)
        # Correct format for math.scatter
        valid_indices = _batch_indices(valid_indices)

        active_mask = math.scatter(self.sample_points, valid_indices, 1, math.concat([[valid_indices.shape[0]], resolution, [1]], axis=-1), duplicates_handling='any')

        mask = math.pad(active_mask, [[0, 0]] + [[1, 1]] * self.rank + [[0, 0]], "constant")

        if isinstance(self.data, (int, float, np.ndarray)):
            values = math.zeros_like(self.sample_points) + self.data
        else:
            values = self.data

        result = []
        ones_1d = math.unstack(math.ones_like(values), axis=-1)[0]
        staggered_shape = [i + 1 for i in resolution]
        dx = box.size / resolution

        dims = range(len(resolution))
        for d in dims:
            staggered_offset = math.stack([(0.5 * dx[i] * ones_1d if i == d else 0.0 * ones_1d) for i in dims], axis=-1)

            indices = math.to_int(math.floor(self.sample_points + staggered_offset))

            valid_indices = math.maximum(0, math.minimum(indices, resolution))
            valid_indices = _batch_indices(valid_indices)

            values_d = math.expand_dims(math.unstack(values, axis=-1)[d], axis=-1)
            result.append(math.scatter(self.sample_points, valid_indices, values_d, [indices.shape[0]] + staggered_shape + [1], duplicates_handling=self.mode))

            d_slice = tuple([(slice(0, -2) if i == d else slice(1,-1)) for i in dims])
            u_slice = tuple([(slice(2, None) if i == d else slice(1,-1)) for i in dims])
            active_mask = math.minimum(mask[(slice(None),) + d_slice + (slice(None),)], active_mask)
            active_mask = math.minimum(mask[(slice(None),) + u_slice + (slice(None),)], active_mask)

        staggered_tensor_prep = unstack_staggered_tensor(math.concat(result, axis=-1))
        grid_values = StaggeredGrid(staggered_tensor_prep)
        # Fix values at boundary of liquids (using StaggeredGrid these might not receive a value, so we replace it with a value inside the liquid)
        grid_values, _ = extrapolate(grid_values, active_mask, voxel_distance=2)
        return grid_values

    @struct.variable()
    def data(self, data):
        assert math.is_tensor(data), data
        rank = math.ndims(data)
        assert rank <= 3
        if rank < 3:
            assert rank in (0, 1)  # Scalar field / vector field
            data = math.expand_dims(data, 0, 3 - rank)
        return math.to_float(data)
    data.override(struct.staticshape, lambda self, data: (self._batch_size, self._point_count, self.component_count) if math.ndims(self.data) > 0 else ())

    @struct.constant(default='mean')
    def mode(self, mode):
        assert mode in ('add', 'mean', 'any')
        return mode

    @struct.variable()
    def sample_points(self, sample_points):
        assert math.is_tensor(sample_points), sample_points
        assert math.ndims(sample_points) == 3, sample_points.shape
        return sample_points
    sample_points.override(struct.staticshape, lambda self, data: (self._batch_size, self._point_count, self.rank))

    @property
    def rank(self):
        return math.staticshape(self.sample_points)[-1]

    @property
    def component_count(self):
        if math.ndims(self.data) == 0:
            return 1
        return math.staticshape(self.data)[-1]

    def unstack(self):
        raise NotImplementedError()

    @property
    def points(self):
        if SAMPLE_POINTS in self.flags or self.sample_points is self.data:
            return self
        return SampledField(self.sample_points, self.sample_points, flags=[SAMPLE_POINTS])

    @property
    def elements(self):
        return Sphere(self.sample_points, radius=0.0)

    def mask(self):
        return self.copied_with(data=1, mode='any', flags=())

    def compatible(self, other_field):
        if not other_field.has_points:
            return True
        if isinstance(other_field, SampledField) and other_field.sample_points is self.sample_points:
            return True
        return False

    def __repr__(self):
        return '%s[%sx(%d), %dD]' % (self.__class__.__name__, self._point_count if self._point_count is not None else '?', self.component_count, self.rank)


def _batch_indices(indices):
    """
Reshapes the indices such that, aside from indices, they also contain batch number.
For example the entry (32, 40) as coordinates of batch 2 will become (2, 32, 40).
Transform shape (b, p, d) to (b, p, d+1) where batch size is b, number of particles is p and number of dimensions is d. 
    """
    batch_size = indices.shape[0]
    out_spatial_rank = len(indices.shape) - 2
    out_spatial_size = math.shape(indices)[1:-1]

    batch_range = math.DYNAMIC_BACKEND.choose_backend(indices).range(batch_size)
    batch_ids = math.reshape(batch_range, [batch_size] + [1] * out_spatial_rank)
    tile_shape = math.pad(out_spatial_size, [[1,0]], constant_values=1)
    batch_ids = math.expand_dims(math.tile(batch_ids, tile_shape), axis=-1)

    return math.concat((batch_ids, indices), axis=-1)


def _distribute_points(density, particles_per_cell=1, distribution='uniform'):
    """
Distribute points according to the distribution specified in density.
    :param density: binary tensor
    :param particles_per_cell: integer
    :param distribution: 'uniform' or 'center'
    :return: tensor of shape (batch_size, point_count, rank)
    """
    assert distribution in ('center', 'uniform')
    index_array = []
    batch_size = math.staticshape(density)[0] if math.staticshape(density)[0] is not None else 1

    for batch in range(batch_size):
        indices = math.where(density[batch, ..., 0] > 0)
        indices = math.to_float(indices)

        temp = []
        for _ in range(particles_per_cell):
            if distribution == 'center':
                temp.append(indices + 0.5)
            elif distribution == 'uniform':
                temp.append(indices + math.random_uniform(math.shape(indices)))
        index_array.append(math.concat(temp, axis=0))
    try:
        index_array = math.stack(index_array)
        return index_array
    except ValueError:
        raise ValueError("all arrays in the batch must have the same number of active cells.")



import warnings

import six

from phi.geom import Geometry, GLOBAL_AXIS_ORDER
from phi.physics.field import Field, mask, ConstantField
from phi import math, struct
from phi.physics import State, Physics, StateDependency

GROW = 'grow'
ADD = 'add'
FIX = 'fix'


@struct.definition()
class FieldEffect(State):

    def __init__(self, field, targets, mode=GROW, bounds=None, tags=('effect',), **kwargs):
        if isinstance(targets, six.string_types):
            targets = [targets]
        tags = tuple(tags) + tuple('%s_effect' % target for target in targets)
        State.__init__(self, **struct.kwargs(locals()))

    @struct.variable()
    def field(self, field):
        assert isinstance(field, Field) or field is None, field
        return field

    @struct.constant()
    def mode(self, mode):
        assert mode in (GROW, ADD, FIX)
        return mode

    @struct.constant()
    def targets(self, targets):
        return tuple(targets)

    @struct.constant()
    def bounds(self, bounds):
        assert isinstance(bounds, Geometry) or bounds is None
        return bounds

    def __repr__(self):
        return '%s(%s to %s)' % (self.mode, self. field, self.targets)


def effect_applied(effect, field, dt):
    effect_field = effect.field.at(field)
    if effect._mode == GROW:
        return field + math.mul(effect_field, dt)
    elif effect._mode == ADD:
        return field + effect_field
    elif effect._mode == FIX:
        assert effect.bounds is not None
        inside = mask([effect.bounds]).at(field)
        return math.where(inside, effect_field, field)
    else:
        raise ValueError('Invalid mode: %s' % effect.mode)


# pylint: disable-msg = invalid-name
def Inflow(geometry, rate=1.0, target='density'): return FieldEffect(mask(geometry, antialias=True) * rate, target, GROW, tags=('inflow', 'effect'))
def Accelerator(geometry, acceleration): return FieldEffect(mask(geometry, antialias=True) * acceleration, ('velocity',), GROW, tags=('fan', 'effect'))
def ConstantVelocity(geometry, velocity): return FieldEffect(ConstantField(velocity), bounds=geometry, targets=('velocity',), mode=FIX, tags=('effect',))
def HeatSource(geometry, rate, name=None): return FieldEffect(mask(geometry, antialias=True) * rate, ('temperature',), GROW, name=name)
def ColdSource(geometry, rate, name=None): return FieldEffect(mask(geometry, antialias=True) * -rate, ('temperature',), GROW, name=name)


def Fan(*args, **kwargs):
    # pylint: disable-msg = invalid-name
    warnings.warn("'Fan' was renamed to 'Accelerator' in version 1.0.2.", DeprecationWarning)
    return Accelerator(*args, **kwargs)


@struct.definition()
class Gravity(State):

    def __init__(self, gravity=-9.81, name='gravity', **kwargs):
        tags = ['gravity']
        State.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def gravity(self, gravity):
        assert gravity is not None
        return gravity

    def __add__(self, other):
        if other is 0:
            return self
        assert isinstance(other, Gravity), type(other)
        if self._batch_size is not None:
            assert self._batch_size == other._batch_size
        # Add gravity
        if math.is_scalar(self.gravity) and math.is_scalar(other.gravity):
            return Gravity(self.gravity + other.gravity)
        else:
            rank = math.staticshape(other.gravity)[-1] if math.is_scalar(self.gravity)\
                else math.staticshape(self.gravity)[-1]
            sum_tensor = gravity_tensor(self, rank) + gravity_tensor(other, rank)
            return Gravity(sum_tensor)

    __radd__ = __add__


def gravity_tensor(gravity, rank):
    if isinstance(gravity, Gravity):
        gravity = gravity.gravity
    if math.is_scalar(gravity):
        gravity = gravity * GLOBAL_AXIS_ORDER.up_vector(rank)
    assert math.staticshape(gravity)[-1] == rank
    return math.to_float(math.expand_dims(gravity, 0, rank + 2 - len(math.staticshape(gravity))))


class FieldPhysics(Physics):

    def __init__(self, fieldname):
        Physics.__init__(self, [StateDependency('effects', fieldname + '_effect', blocking=True)])

    def step(self, field, dt=1.0, effects=()):
        for effect in effects:
            field = effect_applied(effect, field, dt)
        return field.copied_with(age=field.age + dt)



# coding=utf-8
from __future__ import print_function

import inspect
import logging
import numbers
import os
import sys
import threading
import time
import warnings
from os.path import isfile

import numpy as np
import six
from phi import struct
from phi.data.fluidformat import Scene, write_sim_frame
from phi.physics.field import CenteredGrid, Field, StaggeredGrid
from phi.physics.world import StateProxy, world
from phi.viz.plot import PlotlyFigureBuilder

from .control import Action, Control
from .value import (EditableBool, EditableFloat, EditableInt, EditableString, EditableValue)


def synchronized_method(method):
    outer_lock = threading.Lock()
    lock_name = '__' + method.__name__ + '_lock' + '__'

    def sync_method(self, *args, **kws):
        with outer_lock:
            if not hasattr(self, lock_name):
                setattr(self, lock_name, threading.Lock())
            lock = getattr(self, lock_name)
            with lock:
                return method(self, *args, **kws)

    return sync_method


class TimeDependentField(object):

    def __init__(self, name, generator):
        self.name = name
        self.generator = generator
        self.array = None
        self.invalidation_version = -1

    @synchronized_method
    def get(self, invalidation_version):
        if invalidation_version != self.invalidation_version:
            self.array = self.generator()
            self.invalidation_version = invalidation_version
        return self.array


class App(object):

    def __init__(self,
                 name=None,
                 subtitle='',
                 fields=None,
                 stride=None,
                 record_images=False, record_data=False,
                 base_dir='~/phi/data/',
                 recorded_fields=None,
                 summary=None,
                 custom_properties=None,
                 target_scene=None,
                 objects_to_save=None,
                 framerate=None,
                 dt=1.0):
        self.start_time = time.time()
        self.name = name if name is not None else self.__class__.__name__
        self.subtitle = subtitle
        self.summary = summary if summary else name
        if fields:
            self.fields = {name: TimeDependentField(name, generator) for (name, generator) in fields.items()}
        else:
            self.fields = {}
        self.message = None
        self.steps = 0
        self._invalidation_counter = 0
        self._controls = []
        self._actions = []
        self._traits = []
        self.prepared = False
        self.current_action = None
        self._pause = False
        self.detect_fields = 'default'  # False, True, 'default'
        self.world = world
        self.dt = dt
        # Setup directory & Logging
        self.objects_to_save = [self.__class__] if objects_to_save is None else list(objects_to_save)
        self.base_dir = os.path.expanduser(base_dir)
        if not target_scene:
            self.new_scene()
            self.uses_existing_scene = False
        else:
            self.scene = target_scene
            self.uses_existing_scene = True
        if not isfile(self.scene.subpath('info.log')):
            log_file = self.log_file = self.scene.subpath('info.log')
        else:
            index = 2
            while True:
                log_file = self.scene.subpath('info_%d.log' % index)
                if not isfile(log_file):
                    break
                else:
                    index += 1
        # Setup logging
        logFormatter = logging.Formatter('%(message)s (%(levelname)s), %(asctime)sn\n')
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logging.WARNING)
        customLogger = logging.Logger('app', logging.DEBUG)
        fileHandler = logging.FileHandler(log_file)
        fileHandler.setFormatter(logFormatter)
        customLogger.addHandler(fileHandler)
        consoleHandler = logging.StreamHandler(sys.stdout)
        consoleHandler.setFormatter(logFormatter)
        consoleHandler.setLevel(logging.INFO)
        customLogger.addHandler(consoleHandler)
        self.logger = customLogger
        # Recording
        self.record_images = record_images
        self.record_data = record_data
        self.recorded_fields = recorded_fields if recorded_fields is not None else []
        self.rec_all_slices = False
        self.sequence_stride = stride if stride is not None else 1
        self.framerate = framerate if framerate is not None else stride
        self._custom_properties = custom_properties if custom_properties else {}
        self.figures = PlotlyFigureBuilder()
        self.info('App created. Scene directory is %s' % self.scene.path)

    def new_scene(self, count=None):
        if count is None:
            count = 1 if self.world.batch_size is None else self.world.batch_size
        self.scene = Scene.create(self.base_dir, self.scene_summary(), count=count, mkdir=True)

    @property
    def directory(self):
        return self.scene.path

    @property
    def image_dir(self):
        return self.scene.subpath('images')

    def get_image_dir(self):
        return self.scene.subpath('images', create=True)

    def progress(self):
        self.step()
        self.steps += 1
        self.invalidate()

    def invalidate(self):
        self._invalidation_counter += 1

    def step(self):
        world.step(dt=self.dt)

    @property
    def fieldnames(self):
        return sorted(self.fields.keys())

    def get_field(self, fieldname):
        if fieldname not in self.fields:
            raise KeyError('Field %s not declared. Available fields are %s' % (fieldname, self.fields.keys()))
        return self.fields[fieldname].get(self._invalidation_counter)

    def add_field(self, name, value):
        assert not self.prepared, 'Cannot add fields to a prepared model'
        if isinstance(value, StateProxy):
            def current_state():
                return value.state
            generator = current_state
        elif callable(value):
            generator = value
        else:
            assert isinstance(value, (np.ndarray, Field, float, int)), 'Unsupported type for field "%s": %s' % (name, type(value))

            def get_constant():
                return value
            generator = get_constant
        self.fields[name] = TimeDependentField(name, generator)

    @property
    def actions(self):
        return self._actions

    def add_action(self, name, methodcall):
        self._actions.append(Action(name, methodcall, name))

    def run_action(self, action):
        message_before = self.message
        action.method()
        self.invalidate()
        message_after = self.message
        if message_before == message_after:
            if self.message is None or self.message == '':
                self.message = display_name(action.name)
            else:
                self.message += ' | ' + display_name(action.name)

    @property
    def traits(self):
        return self._traits

    def add_trait(self, trait):
        assert not self.prepared, 'Cannot add traits to a prepared model'
        self._traits.append(trait)

    @property
    def controls(self):
        return self._controls

    def prepare(self):
        if self.prepared:
            return
        logging.info('Gathering model data...')
        # Controls
        for name in self.__dict__:
            val = getattr(self, name)
            editable_value = None
            if isinstance(val, EditableValue):
                editable_value = val
                setattr(self, name, val.initial_value)  # Replace EditableValue with initial value
            elif name.startswith('value_'):
                value_name = display_name(name[6:])
                dtype = type(val)
                if dtype == bool:
                    editable_value = EditableBool(value_name, val)
                elif isinstance(val, numbers.Integral):  # Int
                    editable_value = EditableInt(value_name, val)
                elif isinstance(val, numbers.Number):  # Float
                    editable_value = EditableFloat(value_name, val)
                elif isinstance(val, six.string_types):
                    editable_value = EditableString(value_name, val)
            if editable_value:
                self._controls.append(Control(self, name, editable_value))
        # Actions
        for method_name in dir(self):
            if method_name.startswith('action_') and callable(getattr(self, method_name)):
                self._actions.append(Action(display_name(method_name[7:]), getattr(self, method_name), method_name))
        # Default fields
        if len(self.fields) == 0:
            self._add_default_fields()
        # Scene
        self._update_scene_properties()
        source_files_to_save = set()
        for object in self.objects_to_save:
            try:
                source_files_to_save.add(inspect.getabsfile(object))
            except TypeError:
                pass
        for source_file in source_files_to_save:
            self.scene.copy_src(source_file)
        # End
        self.prepared = True
        return self

    def _add_default_fields(self):
        def add_default_field(trace):
            field = trace.value
            if isinstance(field, (CenteredGrid, StaggeredGrid)):
                def field_generator():
                    world_state = self.world.state
                    return trace.find_in(world_state)
                self.add_field(field.name[0].upper() + field.name[1:], field_generator)
            return None
        struct.map(add_default_field, self.world.state, leaf_condition=lambda x: isinstance(x, (CenteredGrid, StaggeredGrid)), trace=True, content_type=struct.INVALID)

    def add_custom_property(self, key, value):
        self._custom_properties[key] = value
        if self.prepared:
            self._update_scene_properties()

    def add_custom_properties(self, dictionary):
        self._custom_properties.update(dictionary)
        if self.prepared:
            self._update_scene_properties()

    def _update_scene_properties(self):
        if self.uses_existing_scene:
            return
        try:
            app_name = os.path.basename(inspect.getfile(self.__class__))
            app_path = inspect.getabsfile(self.__class__)
        except TypeError:
            app_name = app_path = ''
        properties = {
            'instigator': 'App',
            'traits': self.traits,
            'app': str(app_name),
            'app_path': str(app_path),
            'name': self.name,
            'description': self.subtitle,
            'all_fields': self.fieldnames,
            'actions': [action.name for action in self.actions],
            'controls': [{control.name: control.value} for control in self.controls],
            'summary': self.scene_summary(),
            'time_of_writing': self.steps,
            'world': struct.properties_dict(self.world.state)
        }
        properties.update(self.custom_properties())
        self.scene.properties = properties

    def settings_str(self):
        return ''.join([
            ' ' + str(control) for control in self.controls
        ])

    def custom_properties(self):
        return self._custom_properties

    def info(self, message):
        message = str(message)
        self.message = message
        self.logger.info(message)

    def debug(self, message):
        logging.info(message)

    def scene_summary(self):
        return self.summary

    def show(self, *args, **kwargs):
        warnings.warn("Use show(model) instead.", DeprecationWarning, stacklevel=2)
        from phi.viz.display import show
        show(self, *args, **kwargs)

    @property
    def status(self):
        pausing = '/Pausing' if self._pause and self.current_action else ''
        action = self.current_action if self.current_action else 'Idle'
        message = (' - %s' % self.message) if self.message else ''
        return '{}{} ({}){}'.format(action, pausing, self.steps, message)

    def run_step(self, framerate=None, allow_recording=True):
        self.current_action = 'Running'
        starttime = time.time()
        try:
            self.progress()
            if allow_recording and self.steps % self.sequence_stride == 0:
                self.record_frame()
            if framerate is not None:
                duration = time.time() - starttime
                rest = 1.0 / framerate - duration
                if rest > 0:
                    self.current_action = 'Waiting'
                    time.sleep(rest)
        except Exception as e:
            self.info('Error during %s.step() \n %s: %s' % (type(self).__name__, type(e).__name__, e))
            self.logger.exception(e)
        finally:
            self.current_action = None

    def play(self, max_steps=None, callback=None, framerate=None, allow_recording=True, callback_if_aborted=False):
        if framerate is None:
            framerate = self.framerate

        def target():
            self._pause = False
            step_count = 0
            while not self._pause:
                self.run_step(framerate=framerate, allow_recording=allow_recording)
                step_count += 1
                if max_steps and step_count >= max_steps:
                    break
            if callback is not None:
                if not self._pause or callback_if_aborted:
                    callback()

        thread = threading.Thread(target=target)
        thread.start()
        return self

    def pause(self):
        self._pause = True

    @property
    def running(self):
        return self.current_action is not None

    def record_frame(self):
        self.current_action = 'Recording'
        files = []

        if self.record_images:
            os.path.isdir(self.image_dir) or os.makedirs(self.image_dir)
            arrays = [self.get_field(field) for field in self.recorded_fields]
            for name, array in zip(self.recorded_fields, arrays):
                files += self.figures.save_figures(self.image_dir, name, self.steps, array)

        if self.record_data:
            arrays = [self.get_field(field) for field in self.recorded_fields]
            arrays = [a.staggered_tensor() if isinstance(a, StaggeredGrid) else a.data for a in arrays]
            names = [n.lower() for n in self.recorded_fields]
            files += write_sim_frame(self.directory, arrays, names, self.steps)

        if files:
            self.message = 'Frame written to %s' % files
        self.current_action = None

    def benchmark(self, sequence_count):
        self._pause = False
        step_count = 0
        starttime = time.time()
        for i in range(sequence_count):
            self.run_step(framerate=np.inf, allow_recording=False)
            step_count += 1
            if self._pause:
                break
        time_elapsed = time.time() - starttime
        return step_count, time_elapsed

    def config_recording(self, images, data, fields):
        self.record_images = images
        self.record_data = data
        self.recorded_fields = fields


def display_name(python_name):
    n = list(python_name)
    n[0] = n[0].upper()
    for i in range(1, len(n)):
        if n[i] == '_':
            n[i] = ' '
            if len(n) > i + 1:
                n[i + 1] = n[i + 1].upper()
    return ''.join(n)
